@@ -17,8 +17,13 @@ const Mainloop = imports.mainloop;
 const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const Chrome = imports.ui.chrome;
+const CtrlAltTab = imports.ui.ctrlAltTab;
+const EndSessionDialog = imports.ui.endSessionDialog;
+const PolkitAuthenticationAgent = imports.ui.polkitAuthenticationAgent;
 const Environment = imports.ui.environment;
 const ExtensionSystem = imports.ui.extensionSystem;
 const MessageTray = imports.ui.messageTray;
@@ -34,13 +39,16 @@ const ShellDBus = imports.ui.shellDBus;
 const TelepathyClient = imports.ui.telepathyClient;
 const WindowManager = imports.ui.windowManager;
 const Magnifier = imports.ui.magnifier;
+const XdndHandler = imports.ui.xdndHandler;
 const StatusIconDispatcher = imports.ui.statusIconDispatcher;
+const Util = imports.misc.util;
 
 const DEFAULT_BACKGROUND_COLOR = new Clutter.Color();
 DEFAULT_BACKGROUND_COLOR.from_pixel(0x2266bbff);
 
 let chrome = null;
 let panel = null;
+let hotCorners = [];
 let placesManager = null;
 let overview = null;
 let runDialog = null;
@@ -50,15 +58,19 @@ let messageTray = null;
 let notificationDaemon = null;
 let windowAttentionHandler = null;
 let telepathyClient = null;
+let ctrlAltTabManager = null;
 let recorder = null;
 let shellDBusService = null;
 let modalCount = 0;
 let modalActorFocusStack = [];
 let uiGroup = null;
 let magnifier = null;
+let xdndHandler = null;
 let statusIconDispatcher = null;
 let _errorLogStack = [];
 let _startDate;
+let _defaultCssStylesheet = null;
+let _cssStylesheet = null;
 
 let background = null;
 
@@ -74,9 +86,11 @@ function start() {
     global.logError = _logError;
     global.log = _logDebug;
 
+    // Chain up async errors reported from C
+    global.connect('notify-error', function (global, msg, detail) { notifyError(msg, detail); });
+
     Gio.DesktopAppInfo.set_desktop_env('GNOME');
 
-    global.grab_dbus_service();
     shellDBusService = new ShellDBus.GnomeShell();
     // Force a connection now; dbus.js will do this internally
     // if we use its name acquisition stuff but we aren't right
@@ -84,6 +98,10 @@ function start() {
     // back into sync ones.
     DBus.session.flush();
 
+    // Load the calendar server. Note that we are careful about
+    // not loading any events until the user presses the clock
+    global.launch_calendar_server();
+
     Environment.init();
 
     // Ensure ShellWindowTracker and ShellAppUsage are initialized; this will
@@ -97,15 +115,14 @@ function start() {
     Shell.WindowTracker.get_default();
     Shell.AppUsage.get_default();
 
-    // The background color really only matters if there is no desktop
-    // window (say, nautilus) running. We set it mostly so things look good
-    // when we are running inside Xephyr.
+    // The stage is always covered so Clutter doesn't need to clear it; however
+    // the color is used as the default contents for the Mutter root background
+    // actor so set it anyways.
     global.stage.color = DEFAULT_BACKGROUND_COLOR;
+    global.stage.no_clear_hint = true;
 
-    let themeContext = St.ThemeContext.get_for_stage (global.stage);
-    let stylesheetPath = global.datadir + '/theme/gnome-shell.css';
-    let theme = new St.Theme ({ application_stylesheet: stylesheetPath });
-    themeContext.set_theme (theme);
+    _defaultCssStylesheet = global.datadir + '/theme/gnome-shell.css';
+    loadTheme();
 
     let shellwm = global.window_manager;
     shellwm.takeover_keybinding('panel_main_menu');
@@ -119,11 +136,14 @@ function start() {
 
     // Set up stage hierarchy to group all UI actors under one container.
     uiGroup = new Clutter.Group();
+    St.set_ui_root(global.stage, uiGroup);
     global.window_group.reparent(uiGroup);
     global.overlay_group.reparent(uiGroup);
     global.stage.add_actor(uiGroup);
 
     placesManager = new PlaceDisplay.PlacesManager();
+    xdndHandler = new XdndHandler.XdndHandler();
+    ctrlAltTabManager = new CtrlAltTab.CtrlAltTabManager();
     overview = new Overview.Overview();
     chrome = new Chrome.Chrome();
     magnifier = new Magnifier.Magnifier();
@@ -135,6 +155,9 @@ function start() {
     windowAttentionHandler = new WindowAttentionHandler.WindowAttentionHandler();
     telepathyClient = new TelepathyClient.Client();
 
+    overview.init();
+    statusIconDispatcher.start(messageTray.actor);
+
     _startDate = new Date();
 
     let recorderSettings = new Gio.Settings({ schema: 'org.gnome.shell.recorder' });
@@ -161,15 +184,24 @@ function start() {
         }
     });
 
-    background = global.create_root_pixmap_actor();
-    global.stage.add_actor(background);
-    background.lower_bottom();
+    global.screen.override_workspace_layout(Meta.ScreenCorner.TOPLEFT, false, -1, 1);
 
-    global.gdk_screen.connect('monitors-changed', _relayout);
+    // Provide the bus object for gnome-session to
+    // initiate logouts.
+    EndSessionDialog.init();
+
+    // Attempt to become a PolicyKit authentication agent
+    PolkitAuthenticationAgent.init()
+
+    global.screen.connect('monitors-changed', _relayout);
 
     ExtensionSystem.init();
     ExtensionSystem.loadExtensions();
 
+    // Perform initial relayout here
+    _relayout();
+
+    panel.startStatusArea();
     panel.startupAnimation();
 
     let display = global.screen.get_display();
@@ -177,20 +209,233 @@ function start() {
 
     global.stage.connect('captured-event', _globalKeyPressHandler);
 
-    // Perform initial relayout here
-    _relayout();
-
     _log('info', 'loaded at ' + _startDate);
     log('GNOME Shell started at ' + _startDate);
 
-    Mainloop.idle_add(_removeUnusedWorkspaces);
-
     let perfModuleName = GLib.getenv("SHELL_PERF_MODULE");
     if (perfModuleName) {
         let perfOutput = GLib.getenv("SHELL_PERF_OUTPUT");
         let module = eval('imports.perf.' + perfModuleName + ';');
         Scripting.runPerfScript(module, perfOutput);
     }
+
+    global.screen.connect('notify::n-workspaces', _nWorkspacesChanged);
+
+    global.screen.connect('window-entered-monitor', _windowEnteredMonitor);
+    global.screen.connect('window-left-monitor', _windowLeftMonitor);
+
+    _nWorkspacesChanged();
+}
+
+let _workspaces = [];
+let _checkWorkspacesId = 0;
+
+/*
+ * When the last window closed on a workspace is a dialog or splash
+ * screen, we assume that it might be an initial window shown before
+ * the main window of an application, and give the app a grace period
+ * where it can map another window before we remove the workspace.
+ */
+const LAST_WINDOW_GRACE_TIME = 1000;
+
+function _checkWorkspaces() {
+    let i;
+    let emptyWorkspaces = [];
+
+    for (i = 0; i < _workspaces.length; i++) {
+        let lastRemoved = _workspaces[i]._lastRemovedWindow;
+        if (lastRemoved &&
+            (lastRemoved.get_window_type() == Meta.WindowType.SPLASHSCREEN ||
+             lastRemoved.get_window_type() == Meta.WindowType.DIALOG ||
+             lastRemoved.get_window_type() == Meta.WindowType.MODAL_DIALOG))
+                emptyWorkspaces[i] = false;
+        else
+            emptyWorkspaces[i] = true;
+    }
+
+    let windows = global.get_window_actors();
+    for (i = 0; i < windows.length; i++) {
+        let win = windows[i];
+
+        if (win.get_meta_window().is_on_all_workspaces())
+            continue;
+
+        let workspaceIndex = win.get_workspace();
+        emptyWorkspaces[workspaceIndex] = false;
+    }
+
+    // If we don't have an empty workspace at the end, add one
+    if (!emptyWorkspaces[emptyWorkspaces.length -1]) {
+        global.screen.append_new_workspace(false, global.get_current_time());
+        emptyWorkspaces.push(false);
+    }
+
+    let activeWorkspaceIndex = global.screen.get_active_workspace_index();
+    let removingCurrentWorkspace = (emptyWorkspaces[activeWorkspaceIndex] &&
+                                    activeWorkspaceIndex < emptyWorkspaces.length - 1);
+    // Don't enter the overview when removing multiple empty workspaces at startup
+    let showOverview  = (removingCurrentWorkspace &&
+                         !emptyWorkspaces.every(function(x) { return x; }));
+
+    if (removingCurrentWorkspace) {
+        // "Merge" the empty workspace we are removing with the one at the end
+        wm.blockAnimations();
+    }
+
+    // Delete other empty workspaces; do it from the end to avoid index changes
+    for (i = emptyWorkspaces.length - 2; i >= 0; i--) {
+        if (emptyWorkspaces[i])
+            global.screen.remove_workspace(_workspaces[i], global.get_current_time());
+    }
+
+    if (removingCurrentWorkspace) {
+        global.screen.get_workspace_by_index(global.screen.n_workspaces - 1).activate(global.get_current_time());
+        wm.unblockAnimations();
+
+        if (!overview.visible && showOverview)
+            overview.show();
+    }
+
+    _checkWorkspacesId = 0;
+    return false;
+}
+
+function _windowRemoved(workspace, window) {
+    workspace._lastRemovedWindow = window;
+    _queueCheckWorkspaces();
+    Mainloop.timeout_add(LAST_WINDOW_GRACE_TIME, function() {
+        if (workspace._lastRemovedWindow == window) {
+            workspace._lastRemovedWindow = null;
+            _queueCheckWorkspaces();
+        }
+    });
+}
+
+function _windowLeftMonitor(metaScreen, monitorIndex, metaWin) {
+    // If the window left the primary monitor, that
+    // might make that workspace empty
+    if (monitorIndex == global.get_primary_monitor_index())
+        _queueCheckWorkspaces();
+}
+
+function _windowEnteredMonitor(metaScreen, monitorIndex, metaWin) {
+    // If the window entered the primary monitor, that
+    // might make that workspace non-empty
+    if (monitorIndex == global.get_primary_monitor_index())
+        _queueCheckWorkspaces();
+}
+
+function _queueCheckWorkspaces() {
+    if (_checkWorkspacesId == 0)
+        _checkWorkspacesId = Meta.later_add(Meta.LaterType.BEFORE_REDRAW, _checkWorkspaces);
+}
+
+function _nWorkspacesChanged() {
+    let oldNumWorkspaces = _workspaces.length;
+    let newNumWorkspaces = global.screen.n_workspaces;
+
+    if (oldNumWorkspaces == newNumWorkspaces)
+        return false;
+
+    let lostWorkspaces = [];
+    if (newNumWorkspaces > oldNumWorkspaces) {
+        let w;
+
+        // Assume workspaces are only added at the end
+        for (w = oldNumWorkspaces; w < newNumWorkspaces; w++)
+            _workspaces[w] = global.screen.get_workspace_by_index(w);
+
+        for (w = oldNumWorkspaces; w < newNumWorkspaces; w++) {
+            let workspace = _workspaces[w];
+            workspace._windowAddedId = workspace.connect('window-added', _queueCheckWorkspaces);
+            workspace._windowRemovedId = workspace.connect('window-removed', _windowRemoved);
+        }
+
+    } else {
+        // Assume workspaces are only removed sequentially
+        // (e.g. 2,3,4 - not 2,4,7)
+        let removedIndex;
+        let removedNum = oldNumWorkspaces - newNumWorkspaces;
+        for (let w = 0; w < oldNumWorkspaces; w++) {
+            let workspace = global.screen.get_workspace_by_index(w);
+            if (_workspaces[w] != workspace) {
+                removedIndex = w;
+                break;
+            }
+        }
+
+        let lostWorkspaces = _workspaces.splice(removedIndex, removedNum);
+        lostWorkspaces.forEach(function(workspace) {
+                                   workspace.disconnect(workspace._windowAddedId);
+                                   workspace.disconnect(workspace._windowRemovedId);
+                               });
+    }
+
+    _queueCheckWorkspaces();
+
+    return false;
+}
+
+/**
+ * getThemeStylesheet:
+ *
+ * Get the theme CSS file that the shell will load
+ *
+ * Returns: A file path that contains the theme CSS,
+ *          null if using the default
+ */
+function getThemeStylesheet()
+{
+    return _cssStylesheet;
+}
+
+/**
+ * setThemeStylesheet:
+ * @cssStylesheet: A file path that contains the theme CSS,
+ *                  set it to null to use the default
+ *
+ * Set the theme CSS file that the shell will load
+ */
+function setThemeStylesheet(cssStylesheet)
+{
+    _cssStylesheet = cssStylesheet;
+}
+
+/**
+ * loadTheme:
+ *
+ * Reloads the theme CSS file
+ */
+function loadTheme() {
+    let themeContext = St.ThemeContext.get_for_stage (global.stage);
+
+    let cssStylesheet = _defaultCssStylesheet;
+    if (_cssStylesheet != null)
+        cssStylesheet = _cssStylesheet;
+
+    let theme = new St.Theme ({ application_stylesheet: cssStylesheet });
+    themeContext.set_theme (theme);
+}
+
+/**
+ * notifyError:
+ * @msg: An error message
+ * @details: Additional information
+ *
+ * See shell_global_notify_problem().
+ */
+function notifyError(msg, details) {
+    // Also print to stderr so it's logged somewhere
+    if (details)
+        log("error: " + msg + ": " + details);
+    else
+        log("error: " + msg)
+
+    let source = new MessageTray.SystemNotificationSource();
+    messageTray.add(source);
+    let notification = new MessageTray.Notification(source, msg, details);
+    notification.setTransient(true);
+    source.notify(notification);
 }
 
 /**
@@ -235,47 +480,90 @@ function _getAndClearErrorStack() {
 }
 
 function _relayout() {
-    let primary = global.get_primary_monitor();
-    panel.actor.set_position(primary.x, primary.y);
-    panel.actor.set_size(primary.width, Panel.PANEL_HEIGHT);
-    overview.relayout();
+    let monitors = global.get_monitors();
+    // destroy old corners
+    for (let i = 0; i < hotCorners.length; i++)
+        hotCorners[i].destroy();
+    hotCorners = [];
 
-    background.set_size(global.screen_width, global.screen_height);
+
+    let primary = global.get_primary_monitor();
+    for (let i = 0; i < monitors.length; i++) {
+        let monitor = monitors[i];
+        let isPrimary = (monitor.x == primary.x &&
+                         monitor.y == primary.y &&
+                         monitor.width == primary.width &&
+                         monitor.height == primary.height);
+
+        let cornerX = monitor.x;
+        let cornerY = monitor.y;
+        if (St.Widget.get_default_direction() == St.TextDirection.RTL)
+            cornerX += monitor.width;
+
+
+        let haveTopLeftCorner = true;
+
+        /* Check if we have a top left (right for RTL) corner.
+         * I.e. if there is no monitor directly above or to the left(right) */
+        let besideX;
+        if (St.Widget.get_default_direction() == St.TextDirection.RTL)
+            besideX = monitor.x + 1;
+        else
+            besideX = cornerX - 1;
+        let besideY = cornerY;
+        let aboveX = cornerX;
+        let aboveY = cornerY - 1;
+
+        for (let j = 0; j < monitors.length; j++) {
+            if (i == j)
+                continue;
+            let otherMonitor = monitors[j];
+            if (besideX >= otherMonitor.x &&
+                besideX < otherMonitor.x + otherMonitor.width &&
+                besideY >= otherMonitor.y &&
+                besideY < otherMonitor.y + otherMonitor.height) {
+                haveTopLeftCorner = false;
+                break;
+            }
+            if (aboveX >= otherMonitor.x &&
+                aboveX < otherMonitor.x + otherMonitor.width &&
+                aboveY >= otherMonitor.y &&
+                aboveY < otherMonitor.y + otherMonitor.height) {
+                haveTopLeftCorner = false;
+                break;
+            }
+        }
+
+        /* We only want hot corners where there is a natural top-left
+         * corner, and on the primary monitor */
+        if (!isPrimary && !haveTopLeftCorner)
+            continue;
+
+        let corner = new Panel.HotCorner();
+        hotCorners.push(corner);
+        corner.actor.set_position(cornerX, cornerY);
+        if (isPrimary)
+            panel.setHotCorner(corner);
+    }
+
+    panel.relayout();
+    overview.relayout();
 
     // To avoid updating the position and size of the workspaces
     // in the overview, we just hide the overview. The positions
-    // will be updated when it is next shown. We do the same for
-    // the calendar popdown.
+    // will be updated when it is next shown.
     overview.hide();
-    panel.hideCalendar();
 }
 
-// metacity-clutter currently uses the same prefs as plain metacity,
-// which probably means we'll be starting out with multiple workspaces;
-// remove any unused ones. (We do this from an idle handler, because
-// global.get_window_actors() still returns NULL at the point when start()
-// is called.)
-function _removeUnusedWorkspaces() {
+function isWindowActorDisplayedOnWorkspace(win, workspaceIndex) {
+    return win.get_workspace() == workspaceIndex ||
+        (win.get_meta_window() && win.get_meta_window().is_on_all_workspaces());
+}
 
-    let windows = global.get_window_actors();
-    let maxWorkspace = 0;
-    for (let i = 0; i < windows.length; i++) {
-        let win = windows[i];
-
-        if (!win.get_meta_window().is_on_all_workspaces() &&
-            win.get_workspace() > maxWorkspace) {
-            maxWorkspace = win.get_workspace();
-        }
-    }
-    let screen = global.screen;
-    if (screen.n_workspaces > maxWorkspace) {
-        for (let w = screen.n_workspaces - 1; w > maxWorkspace; w--) {
-            let workspace = screen.get_workspace_by_index(w);
-            screen.remove_workspace(workspace, 0);
-        }
-    }
-
-    return false;
+function getWindowActorsForWorkspace(workspaceIndex) {
+    return global.get_window_actors().filter(function (win) {
+        return isWindowActorDisplayedOnWorkspace(win, workspaceIndex);
+    });
 }
 
 // This function encapsulates hacks to make certain global keybindings
@@ -283,57 +571,67 @@ function _removeUnusedWorkspaces() {
 // are disabled with a global grab. (When there is a global grab, then
 // all key events will be delivered to the stage, so ::captured-event
 // on the stage can be used for global keybindings.)
-//
-// We expect to need to conditionally enable just a few keybindings
-// depending on circumstance; the main hackiness here is that we are
-// assuming that keybindings have their default values; really we
-// should be asking Mutter to resolve the key into an action and then
-// base our handling based on the action.
 function _globalKeyPressHandler(actor, event) {
     if (modalCount == 0)
         return false;
-    if (event.type() != Clutter.EventType.KEY_RELEASE)
+    if (event.type() != Clutter.EventType.KEY_PRESS)
         return false;
 
     let symbol = event.get_key_symbol();
     let keyCode = event.get_key_code();
     let modifierState = Shell.get_event_state(event);
-    // Check the overview key first, this isn't a Meta.KeyBindingAction yet
-    if (symbol == Clutter.Super_L || symbol == Clutter.Super_R) {
-        // The super key is the default for triggering the overview, and should
-        // get us out of the overview when we are already in it.
-        if (overview.visible)
-            overview.hide();
 
+    let display = global.screen.get_display();
+    // This relies on the fact that Clutter.ModifierType is the same as Gdk.ModifierType
+    let action = display.get_keybinding_action(keyCode, modifierState);
+
+    // The screenshot action should always be available (even if a
+    // modal dialog is present)
+    if (action == Meta.KeyBindingAction.COMMAND_SCREENSHOT) {
+        let gconf = GConf.Client.get_default();
+        let command = gconf.get_string('/apps/metacity/keybinding_commands/command_screenshot');
+        if (command != null && command != '')
+            Util.spawnCommandLine(command);
         return true;
     }
 
-    // Whitelist some of the Metacity actions
-    let display = global.screen.get_display();
-    let activeWorkspaceIndex = global.screen.get_active_workspace_index();
+    // Other bindings are only available when the overview is up and
+    // no modal dialog is present.
+    if (!overview.visible || modalCount > 1)
+        return false;
+
+    // This isn't a Meta.KeyBindingAction yet
+    if (symbol == Clutter.Super_L || symbol == Clutter.Super_R) {
+        overview.hide();
+        return true;
+    }
 
-    // This relies on the fact that Clutter.ModifierType is the same as Gdk.ModifierType
-    let action = display.get_keybinding_action(keyCode, modifierState);
     switch (action) {
-        case Meta.KeyBindingAction.COMMAND_SCREENSHOT:
-            let gconf = GConf.Client.get_default();
-            let command = gconf.get_string('/apps/metacity/keybinding_commands/command_screenshot');
-            if (command != null && command != '') {
-                let [ok, len, args] = GLib.shell_parse_argv(command);
-                let p = new Shell.Process({'args' : args});
-                p.run();
-            }
+        // left/right would effectively act as synonyms for up/down if we enabled them;
+        // but that could be considered confusing; we also disable them in the main view.
+        //
+        // case Meta.KeyBindingAction.WORKSPACE_LEFT:
+        //     wm.actionMoveWorkspaceLeft();
+        //     return true;
+        // case Meta.KeyBindingAction.WORKSPACE_RIGHT:
+        //     wm.actionMoveWorkspaceRight();
+        //     return true;
+        case Meta.KeyBindingAction.WORKSPACE_UP:
+            wm.actionMoveWorkspaceUp();
             return true;
-        case Meta.KeyBindingAction.WORKSPACE_LEFT:
-            wm.actionMoveWorkspaceLeft();
-            return true;
-        case Meta.KeyBindingAction.WORKSPACE_RIGHT:
-            wm.actionMoveWorkspaceRight();
+        case Meta.KeyBindingAction.WORKSPACE_DOWN:
+            wm.actionMoveWorkspaceDown();
             return true;
         case Meta.KeyBindingAction.PANEL_RUN_DIALOG:
         case Meta.KeyBindingAction.COMMAND_2:
             getRunDialog().open();
             return true;
+        case Meta.KeyBindingAction.PANEL_MAIN_MENU:
+            overview.hide();
+            return true;
+        case Meta.KeyBindingAction.SWITCH_PANELS:
+            ctrlAltTabManager.popup(modifierState & Clutter.ModifierType.SHIFT_MASK);
+            return true;
     }
 
     return false;
@@ -341,10 +639,8 @@ function _globalKeyPressHandler(actor, event) {
 
 function _findModal(actor) {
     for (let i = 0; i < modalActorFocusStack.length; i++) {
-        let [stackActor, stackFocus] = modalActorFocusStack[i];
-        if (stackActor == actor) {
+        if (modalActorFocusStack[i].actor == actor)
             return i;
-        }
     }
     return -1;
 }
@@ -352,21 +648,29 @@ function _findModal(actor) {
 /**
  * pushModal:
  * @actor: #ClutterActor which will be given keyboard focus
+ * @timestamp: optional timestamp
  *
  * Ensure we are in a mode where all keyboard and mouse input goes to
- * the stage.  Multiple calls to this function act in a stacking fashion;
- * the effect will be undone when an equal number of popModal() invocations
- * have been made.
+ * the stage, and focus @actor. Multiple calls to this function act in
+ * a stacking fashion; the effect will be undone when an equal number
+ * of popModal() invocations have been made.
  *
- * Next, record the current Clutter keyboard focus on a stack.  If the modal stack
- * returns to this actor, reset the focus to the actor which was focused
- * at the time pushModal() was invoked.
+ * Next, record the current Clutter keyboard focus on a stack. If the
+ * modal stack returns to this actor, reset the focus to the actor
+ * which was focused at the time pushModal() was invoked.
+ *
+ * @timestamp is optionally used to associate the call with a specific user
+ * initiated event.  If not provided then the value of
+ * global.get_current_time() is assumed.
  *
  * Returns: true iff we successfully acquired a grab or already had one
  */
-function pushModal(actor) {
+function pushModal(actor, timestamp) {
+    if (timestamp == undefined)
+        timestamp = global.get_current_time();
+
     if (modalCount == 0) {
-        if (!global.begin_modal(global.get_current_time())) {
+        if (!global.begin_modal(timestamp)) {
             log('pushModal: invocation of begin_modal failed');
             return false;
         }
@@ -375,52 +679,80 @@ function pushModal(actor) {
     global.set_stage_input_mode(Shell.StageInputMode.FULLSCREEN);
 
     modalCount += 1;
-    actor.connect('destroy', function() {
+    let actorDestroyId = actor.connect('destroy', function() {
         let index = _findModal(actor);
         if (index >= 0)
             modalActorFocusStack.splice(index, 1);
     });
     let curFocus = global.stage.get_key_focus();
+    let curFocusDestroyId;
     if (curFocus != null) {
-        curFocus.connect('destroy', function() {
+        curFocusDestroyId = curFocus.connect('destroy', function() {
             let index = _findModal(actor);
             if (index >= 0)
-                modalActorFocusStack[index][1] = null;
+                modalActorFocusStack[index].actor = null;
         });
     }
-    modalActorFocusStack.push([actor, curFocus]);
+    modalActorFocusStack.push({ actor: actor,
+                                focus: curFocus,
+                                destroyId: actorDestroyId,
+                                focusDestroyId: curFocusDestroyId });
 
-    global.stage.set_key_focus(null);
+    global.stage.set_key_focus(actor);
     return true;
 }
 
 /**
  * popModal:
  * @actor: #ClutterActor passed to original invocation of pushModal().
+ * @timestamp: optional timestamp
  *
  * Reverse the effect of pushModal().  If this invocation is undoing
  * the topmost invocation, then the focus will be restored to the
  * previous focus at the time when pushModal() was invoked.
+ *
+ * @timestamp is optionally used to associate the call with a specific user
+ * initiated event.  If not provided then the value of
+ * global.get_current_time() is assumed.
  */
-function popModal(actor) {
-    modalCount -= 1;
+function popModal(actor, timestamp) {
+    if (timestamp == undefined)
+        timestamp = global.get_current_time();
+
     let focusIndex = _findModal(actor);
-    if (focusIndex >= 0) {
-        if (focusIndex == modalActorFocusStack.length - 1) {
-            let [stackActor, stackFocus] = modalActorFocusStack[focusIndex];
-            global.stage.set_key_focus(stackFocus);
-        } else {
-            // Remove from the middle, shift the focus chain up
-            for (let i = focusIndex; i < modalActorFocusStack.length - 1; i++) {
-                modalActorFocusStack[i + 1][1] = modalActorFocusStack[i][1];
-            }
-        }
-        modalActorFocusStack.splice(focusIndex, 1);
+    if (focusIndex < 0) {
+        global.stage.set_key_focus(null);
+        global.end_modal(timestamp);
+        global.set_stage_input_mode(Shell.StageInputMode.NORMAL);
+
+        throw new Error('incorrect pop');
     }
+
+    modalCount -= 1;
+
+    let record = modalActorFocusStack[focusIndex];
+    record.actor.disconnect(record.destroyId);
+
+    if (focusIndex == modalActorFocusStack.length - 1) {
+        if (record.focus)
+            record.focus.disconnect(record.focusDestroyId);
+        global.stage.set_key_focus(record.focus);
+    } else {
+        let t = modalActorFocusStack[modalActorFocusStack.length - 1];
+        if (t.focus)
+            t.focus.disconnect(t.focusDestroyId);
+        // Remove from the middle, shift the focus chain up
+        for (let i = modalActorFocusStack.length - 1; i > focusIndex; i--) {
+            modalActorFocusStack[i].focus = modalActorFocusStack[i - 1].focus;
+            modalActorFocusStack[i].focusDestroyId = modalActorFocusStack[i - 1].focusDestroyId;
+        }
+    }
+    modalActorFocusStack.splice(focusIndex, 1);
+
     if (modalCount > 0)
         return;
 
-    global.end_modal(global.get_current_time());
+    global.end_modal(timestamp);
     global.set_stage_input_mode(Shell.StageInputMode.NORMAL);
 }
 
