@@ -1,6 +1,8 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const Gdm = imports.gi.Gdm;
+const DBus = imports.dbus;
+const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
 const Lang = imports.lang;
 const Shell = imports.gi.Shell;
@@ -16,6 +18,21 @@ const PanelMenu = imports.ui.panelMenu;
 const PopupMenu = imports.ui.popupMenu;
 const Util = imports.misc.util;
 
+const BUS_NAME = 'org.gnome.ScreenSaver';
+const OBJECT_PATH = '/org/gnome/ScreenSaver';
+
+const LOCKDOWN_SCHEMA = 'org.gnome.desktop.lockdown';
+const DISABLE_USER_SWITCH_KEY = 'disable-user-switching';
+const DISABLE_LOCK_SCREEN_KEY = 'disable-lock-screen';
+const DISABLE_LOG_OUT_KEY = 'disable-log-out';
+
+const ScreenSaverInterface = {
+    name: BUS_NAME,
+    methods: [ { name: 'Lock', inSignature: '' } ]
+};
+
+let ScreenSaverProxy = DBus.makeProxyClass(ScreenSaverInterface);
+
 // Adapted from gdm/gui/user-switch-applet/applet.c
 //
 // Copyright (C) 2004-2005 James M. Cape <jcape@ignore-your.tv>.
@@ -33,17 +50,20 @@ StatusMenuButton.prototype = {
         let box = new St.BoxLayout({ name: 'panelStatusMenu' });
         this.actor.set_child(box);
 
+        this._lockdownSettings = new Gio.Settings({ schema: LOCKDOWN_SCHEMA });
+
         this._gdm = Gdm.UserManager.ref_default();
         this._gdm.queue_load();
 
         this._user = this._gdm.get_user(GLib.get_user_name());
         this._presence = new GnomeSession.Presence();
         this._presenceItems = {};
+        this._session = new GnomeSession.SessionManager();
 
         this._account_mgr = Tp.AccountManager.dup()
 
         this._upClient = new UPowerGlib.Client();
-
+        this._screenSaverProxy = new ScreenSaverProxy(DBus.session, BUS_NAME, OBJECT_PATH);
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
         this._iconBox = new St.Bin();
@@ -67,6 +87,15 @@ StatusMenuButton.prototype = {
         this._gdm.connect('notify::is-loaded', Lang.bind(this, this._updateSwitchUser));
         this._gdm.connect('user-added', Lang.bind(this, this._updateSwitchUser));
         this._gdm.connect('user-removed', Lang.bind(this, this._updateSwitchUser));
+        this._lockdownSettings.connect('changed::' + DISABLE_USER_SWITCH_KEY,
+                                       Lang.bind(this, this._updateSwitchUser));
+        this._lockdownSettings.connect('changed::' + DISABLE_LOG_OUT_KEY,
+                                       Lang.bind(this, this._updateLogout));
+        this._lockdownSettings.connect('changed::' + DISABLE_LOCK_SCREEN_KEY,
+                                       Lang.bind(this, this._updateLockScreen));
+        this._updateSwitchUser();
+        this._updateLogout();
+        this._updateLockScreen();
 
         this._upClient.connect('notify::can-suspend', Lang.bind(this, this._updateSuspendOrPowerOff));
     },
@@ -83,11 +112,41 @@ StatusMenuButton.prototype = {
           this._name.set_text("");
     },
 
+    _updateSessionSeparator: function() {
+        let showSeparator = this._loginScreenItem.actor.visible ||
+                            this._logoutItem.actor.visible ||
+                            this._lockScreenItem.actor.visible;
+        if (showSeparator)
+            this._sessionSeparator.actor.show();
+        else
+            this._sessionSeparator.actor.hide();
+    },
+
     _updateSwitchUser: function() {
-        if (this._gdm.can_switch ())
+        let allowSwitch = !this._lockdownSettings.get_boolean(DISABLE_USER_SWITCH_KEY);
+        if (allowSwitch && this._gdm.can_switch ())
             this._loginScreenItem.actor.show();
         else
             this._loginScreenItem.actor.hide();
+        this._updateSessionSeparator();
+    },
+
+    _updateLogout: function() {
+        let allowLogout = !this._lockdownSettings.get_boolean(DISABLE_LOG_OUT_KEY);
+        if (allowLogout)
+            this._logoutItem.actor.show();
+        else
+            this._logoutItem.actor.hide();
+        this._updateSessionSeparator();
+    },
+
+    _updateLockScreen: function() {
+        let allowLockScreen = !this._lockdownSettings.get_boolean(DISABLE_LOCK_SCREEN_KEY);
+        if (allowLockScreen)
+            this._lockScreenItem.actor.show();
+        else
+            this._lockScreenItem.actor.hide();
+        this._updateSessionSeparator();
     },
 
     _updateSuspendOrPowerOff: function() {
@@ -149,6 +208,7 @@ StatusMenuButton.prototype = {
         item = new PopupMenu.PopupMenuItem(_("Lock Screen"));
         item.connect('activate', Lang.bind(this, this._onLockScreenActivate));
         this.menu.addMenuItem(item);
+        this._lockScreenItem = item;
 
         item = new PopupMenu.PopupMenuItem(_("Switch User"));
         item.connect('activate', Lang.bind(this, this._onLoginScreenActivate));
@@ -158,9 +218,11 @@ StatusMenuButton.prototype = {
         item = new PopupMenu.PopupMenuItem(_("Log Out..."));
         item.connect('activate', Lang.bind(this, this._onQuitSessionActivate));
         this.menu.addMenuItem(item);
+        this._logoutItem = item;
 
         item = new PopupMenu.PopupSeparatorMenuItem();
         this.menu.addMenuItem(item);
+        this._sessionSeparator = item;
 
         item = new PopupMenu.PopupAlternatingMenuItem(_("Suspend"),
                                                       _("Power Off..."));
@@ -178,17 +240,19 @@ StatusMenuButton.prototype = {
 
     _onMyAccountActivate: function() {
         Main.overview.hide();
-        Util.spawnDesktop('gnome-user-accounts-panel');
+        let app = Shell.AppSystem.get_default().get_app('gnome-user-accounts-panel.desktop');
+        app.activate(-1);
     },
 
     _onPreferencesActivate: function() {
         Main.overview.hide();
-        Util.spawnDesktop('gnome-control-center');
+        let app = Shell.AppSystem.get_default().get_app('gnome-control-center.desktop');
+        app.activate(-1);
     },
 
     _onLockScreenActivate: function() {
         Main.overview.hide();
-        Util.spawn(['gnome-screensaver-command', '--lock']);
+        this._screenSaverProxy.LockRemote();
     },
 
     _onLoginScreenActivate: function() {
@@ -199,7 +263,7 @@ StatusMenuButton.prototype = {
 
     _onQuitSessionActivate: function() {
         Main.overview.hide();
-        Util.spawn(['gnome-session-quit', '--logout']);
+        this._session.LogoutRemote(0);
     },
 
     _onSuspendOrPowerOffActivate: function() {
@@ -207,9 +271,11 @@ StatusMenuButton.prototype = {
 
         if (this._haveSuspend &&
             this._suspendOrPowerOffItem.state == PopupMenu.PopupAlternatingMenuItemState.DEFAULT) {
-            this._upClient.suspend_sync(null);
+            this._screenSaverProxy.LockRemote(Lang.bind(this, function() {
+                this._upClient.suspend_sync(null);
+            }));
         } else {
-            Util.spawn(['gnome-session-quit', '--power-off']);
+            this._session.ShutdownRemote();
         }
     },
 
