@@ -10,10 +10,11 @@
 #include <gtk/gtk.h>
 #include <clutter/clutter.h>
 #include <glib/gi18n.h>
+#include <meta/display.h>
 
 #include "shell-app-private.h"
+#include "shell-window-tracker-private.h"
 #include "shell-global.h"
-#include "display.h"
 #include "st.h"
 
 #define GMENU_I_KNOW_THIS_IS_UNSTABLE
@@ -22,10 +23,10 @@
 /* Vendor prefixes are something that can be preprended to a .desktop
  * file name.  Undo this.
  */
-static const char*const known_vendor_prefixes[] = { "gnome",
-                                                    "fedora",
-                                                    "mozilla",
-                                                    NULL };
+static const char*const vendor_prefixes[] = { "gnome-",
+                                              "fedora-",
+                                              "mozilla-",
+                                              NULL };
 
 enum {
    PROP_0,
@@ -48,12 +49,14 @@ struct _ShellAppSystemPrivate {
 
   GSList *cached_flattened_apps; /* ShellAppInfo */
   GSList *cached_settings; /* ShellAppInfo */
+  GSList *known_vendor_prefixes;
 
   gint app_monitor_id;
 
   guint app_change_timeout_id;
 };
 
+static char *shell_app_info_get_prefix (ShellAppInfo *info);
 static void shell_app_system_finalize (GObject *object);
 static gboolean on_tree_changed (gpointer user_data);
 static void on_tree_changed_cb (GMenuTree *tree, gpointer user_data);
@@ -158,20 +161,6 @@ shell_app_info_new_from_window (MetaWindow *window)
   return info;
 }
 
-static ShellAppInfo *
-shell_app_info_new_from_keyfile_take_ownership (GKeyFile   *keyfile,
-                                                const char *path)
-{
-  ShellAppInfo *info;
-
-  info = g_slice_alloc0 (sizeof (ShellAppInfo));
-  info->type = SHELL_APP_INFO_TYPE_DESKTOP_FILE;
-  info->refcount = 1;
-  info->keyfile = keyfile;
-  info->keyfile_path = g_strdup (path);
-  return info;
-}
-
 static void shell_app_system_class_init(ShellAppSystemClass *klass)
 {
   GObjectClass *gobject_class = (GObjectClass *)klass;
@@ -239,6 +228,11 @@ shell_app_system_finalize (GObject *object)
   g_slist_foreach (priv->cached_flattened_apps, (GFunc)shell_app_info_unref, NULL);
   g_slist_free (priv->cached_flattened_apps);
   priv->cached_flattened_apps = NULL;
+
+  g_slist_foreach (priv->known_vendor_prefixes, (GFunc)g_free, NULL);
+  g_slist_free (priv->known_vendor_prefixes);
+  priv->known_vendor_prefixes = NULL;
+
   g_slist_foreach (priv->cached_settings, (GFunc)shell_app_info_unref, NULL);
   g_slist_free (priv->cached_settings);
   priv->cached_settings = NULL;
@@ -322,10 +316,20 @@ cache_by_id (ShellAppSystem *self, GSList *apps)
   for (iter = apps; iter; iter = iter->next)
     {
       ShellAppInfo *info = iter->data;
+      const char *id = shell_app_info_get_id (info);
+      char *prefix = shell_app_info_get_prefix (info);
+
       shell_app_info_ref (info);
       /* the name is owned by the info itself */
-      g_hash_table_replace (self->priv->app_id_to_info, (char*)shell_app_info_get_id (info),
-                            info);
+
+      if (prefix
+          && !g_slist_find_custom (self->priv->known_vendor_prefixes, prefix,
+                                   (GCompareFunc)g_strcmp0))
+        self->priv->known_vendor_prefixes = g_slist_append (self->priv->known_vendor_prefixes,
+                                                            prefix);
+      else
+        g_free (prefix);
+      g_hash_table_replace (self->priv->app_id_to_info, (char*)id, info);
     }
 }
 
@@ -334,6 +338,10 @@ reread_menus (ShellAppSystem *self)
 {
   GHashTable *unique = g_hash_table_new (g_str_hash, g_str_equal);
 
+  g_slist_foreach (self->priv->known_vendor_prefixes, (GFunc)g_free, NULL);
+  g_slist_free (self->priv->known_vendor_prefixes);
+  self->priv->known_vendor_prefixes = NULL;
+
   reread_entries (self, &(self->priv->cached_flattened_apps), unique, self->priv->apps_tree);
   g_hash_table_remove_all (unique);
   reread_entries (self, &(self->priv->cached_settings), unique, self->priv->settings_tree);
@@ -575,44 +583,6 @@ _shell_app_system_register_app (ShellAppSystem   *self,
   g_object_weak_ref (G_OBJECT (app), shell_app_system_on_app_weakref, ref);
 }
 
-ShellAppInfo *
-shell_app_system_load_from_desktop_file (ShellAppSystem   *system,
-                                         const char       *filename,
-                                         GError          **error)
-{
-  ShellAppInfo *appinfo;
-  GKeyFile *keyfile;
-  char *full_path = NULL;
-  gboolean success;
-
-  keyfile = g_key_file_new ();
-
-  if (strchr (filename, '/') != NULL)
-    {
-      success = g_key_file_load_from_file (keyfile, filename, G_KEY_FILE_NONE, error);
-      full_path = g_strdup (filename);
-    }
-  else
-    {
-      char *app_path = g_build_filename ("applications", filename, NULL);
-      success = g_key_file_load_from_data_dirs (keyfile, app_path, &full_path,
-                                                G_KEY_FILE_NONE, error);
-      g_free (app_path);
-    }
-
-  if (!success)
-    {
-      g_key_file_free (keyfile);
-      g_free (full_path);
-      return NULL;
-    }
-
-  appinfo = shell_app_info_new_from_keyfile_take_ownership (keyfile, full_path);
-  g_free (full_path);
-
-  return appinfo;
-}
-
 /**
  * shell_app_system_create_from_window:
  *
@@ -644,16 +614,13 @@ shell_app_system_lookup_heuristic_basename (ShellAppSystem *system,
                                             const char *name)
 {
   ShellApp *result;
-  char **vendor_prefixes;
-
+  GSList *prefix;
   result = shell_app_system_get_app (system, name);
   if (result != NULL)
     return result;
-
-  for (vendor_prefixes = (char**)known_vendor_prefixes;
-       *vendor_prefixes; vendor_prefixes++)
+  for (prefix = system->priv->known_vendor_prefixes; prefix; prefix = g_slist_next (prefix))
     {
-      char *tmpid = g_strjoin (NULL, *vendor_prefixes, "-", name, NULL);
+      char *tmpid = g_strconcat ((char*)prefix->data, name, NULL);
       result = shell_app_system_get_app (system, tmpid);
       g_free (tmpid);
       if (result != NULL)
@@ -992,6 +959,115 @@ shell_app_info_get_id (ShellAppInfo *info)
   return NULL;
 }
 
+static char *
+shell_app_info_get_prefix (ShellAppInfo *info)
+{
+  char *prefix = NULL, *file_prefix = NULL;
+  const char *id;
+  GFile *file;
+  char *name;
+  int i = 0;
+
+  if (info->type != SHELL_APP_INFO_TYPE_ENTRY)
+    return NULL;
+
+  id = gmenu_tree_entry_get_desktop_file_id ((GMenuTreeEntry*)info->entry);
+  file = g_file_new_for_path (gmenu_tree_entry_get_desktop_file_path ((GMenuTreeEntry*)info->entry));
+  name = g_file_get_basename (file);
+
+  if (!name)
+    {
+      g_object_unref (file);
+      return NULL;
+    }
+  for (i = 0; vendor_prefixes[i]; i++)
+    {
+      if (g_str_has_prefix (name, vendor_prefixes[i]))
+        {
+          file_prefix = g_strdup (vendor_prefixes[i]);
+          break;
+        }
+    }
+
+  while (strcmp (name, id) != 0)
+    {
+      char *t;
+      char *pname;
+      GFile *parent = g_file_get_parent (file);
+
+      if (!parent)
+        {
+          g_warn_if_reached ();
+          break;
+        }
+
+      pname = g_file_get_basename (parent);
+      if (!pname)
+        {
+          g_object_unref (parent);
+          break;
+        }
+      if (!g_strstr_len (id, -1, pname))
+        {
+          /* handle <LegacyDir prefix="..."> */
+          char *t;
+          size_t name_len = strlen (name);
+          size_t id_len = strlen (id);
+          char *t_id = g_strdup (id);
+
+          t_id[id_len - name_len] = '\0';
+          t = g_strdup(t_id);
+          g_free (prefix);
+          g_free (t_id);
+          g_free (name);
+          name = g_strdup (id);
+          prefix = t;
+
+          g_object_unref (file);
+          file = parent;
+          g_free (pname);
+          g_free (file_prefix);
+          file_prefix = NULL;
+          break;
+        }
+
+      t = g_strconcat (pname, "-", name, NULL);
+      g_free (name);
+      name = t;
+
+      t = g_strconcat (pname, "-", prefix, NULL);
+      g_free (prefix);
+      prefix = t;
+
+      g_object_unref (file);
+      file = parent;
+      g_free (pname);
+    }
+
+  if (file)
+    g_object_unref (file);
+
+  if (strcmp (name, id) == 0)
+    {
+      g_free (name);
+      if (file_prefix && !prefix)
+        return file_prefix;
+      if (file_prefix)
+        {
+          char *t = g_strconcat (prefix, "-", file_prefix, NULL);
+          g_free (prefix);
+          g_free (file_prefix);
+          prefix = t;
+        }
+      return prefix;
+    }
+
+  g_free (name);
+  g_free (prefix);
+  g_free (file_prefix);
+  g_return_val_if_reached (NULL);
+}
+
 #define DESKTOP_ENTRY_GROUP "Desktop Entry"
 
 char *
@@ -1259,21 +1335,27 @@ shell_app_info_create_icon_texture (ShellAppInfo *info, float size)
   GIcon *icon;
   ClutterActor *ret;
 
+  ret = NULL;
+
   if (info->type == SHELL_APP_INFO_TYPE_WINDOW)
     {
-      return st_texture_cache_bind_pixbuf_property (st_texture_cache_get_default (),
-                                                    G_OBJECT (info->window),
-                                                    "icon");
-    }
-
-  icon = shell_app_info_get_icon (info);
-  if (icon == NULL)
-    {
-      ret = clutter_texture_new ();
-      g_object_set (ret, "opacity", 0, "width", size, "height", size, NULL);
+      ret = st_texture_cache_bind_pixbuf_property (st_texture_cache_get_default (),
+                                                   G_OBJECT (info->window),
+                                                   "icon");
     }
   else
     {
+      icon = shell_app_info_get_icon (info);
+      if (icon != NULL)
+        {
+          ret = st_texture_cache_load_gicon (st_texture_cache_get_default (), NULL, icon, (int)size);
+          g_object_unref (icon);
+        }
+    }
+
+  if (ret == NULL)
+    {
+      icon = g_themed_icon_new ("application-x-executable");
       ret = st_texture_cache_load_gicon (st_texture_cache_get_default (), NULL, icon, (int)size);
       g_object_unref (icon);
     }
@@ -1296,6 +1378,24 @@ shell_app_info_get_source_window (ShellAppInfo *info)
   return NULL;
 }
 
+static void
+_gather_pid_callback (GDesktopAppInfo   *gapp,
+                      GPid               pid,
+                      gpointer           data)
+{
+  ShellApp *app;
+  ShellWindowTracker *tracker;
+
+  g_return_if_fail (data != NULL);
+
+  app = SHELL_APP (data);
+  tracker = shell_window_tracker_get_default ();
+
+  _shell_window_tracker_add_child_process_app (tracker,
+                                               pid,
+                                               app);
+}
+
 /**
  * shell_app_info_launch_full:
  * @timestamp: Event timestamp, or 0 for current event timestamp
@@ -1312,12 +1412,12 @@ shell_app_info_launch_full (ShellAppInfo *info,
                             char        **startup_id,
                             GError      **error)
 {
+  ShellApp *shell_app;
   GDesktopAppInfo *gapp;
   GdkAppLaunchContext *context;
   gboolean ret;
   ShellGlobal *global;
   MetaScreen *screen;
-  MetaDisplay *display;
 
   if (startup_id)
     *startup_id = NULL;
@@ -1353,7 +1453,6 @@ shell_app_info_launch_full (ShellAppInfo *info,
 
   global = shell_global_get ();
   screen = shell_global_get_screen (global);
-  display = meta_screen_get_display (screen);
 
   if (timestamp == 0)
     timestamp = clutter_get_current_event_time ();
@@ -1365,7 +1464,26 @@ shell_app_info_launch_full (ShellAppInfo *info,
   gdk_app_launch_context_set_timestamp (context, timestamp);
   gdk_app_launch_context_set_desktop (context, workspace);
 
-  ret = g_app_info_launch (G_APP_INFO (gapp), uris, (GAppLaunchContext*) context, error);
+  shell_app = shell_app_system_get_app (shell_app_system_get_default (),
+                                        shell_app_info_get_id (info));
+
+  /* In the case where we know an app, we handle reaping the child internally,
+   * in the window tracker.
+   */
+  if (shell_app != NULL)
+    ret = g_desktop_app_info_launch_uris_as_manager (gapp, uris,
+                                                     G_APP_LAUNCH_CONTEXT (context),
+                                                     G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
+                                                     NULL, NULL,
+                                                     _gather_pid_callback, shell_app,
+                                                     error);
+  else
+    ret = g_desktop_app_info_launch_uris_as_manager (gapp, uris,
+                                                     G_APP_LAUNCH_CONTEXT (context),
+                                                     G_SPAWN_SEARCH_PATH,
+                                                     NULL, NULL,
+                                                     NULL, NULL,
+                                                     error);
 
   g_object_unref (G_OBJECT (gapp));
 
