@@ -12,16 +12,15 @@
 #include <glib.h>
 #include <gio/gio.h>
 #include <dbus/dbus-glib.h>
+#include <meta/display.h>
+#include <meta/group.h>
+#include <meta/window.h>
 
 #include "shell-app-usage.h"
 #include "shell-window-tracker.h"
 #include "shell-global.h"
 #include "shell-marshal.h"
 
-#include "display.h"
-#include "window.h"
-#include "group.h"
-
 /* This file includes modified code from
  * desktop-data-engine/engine-dbus/hippo-application-monitor.c
  * in the functions collecting application usage data.
@@ -324,7 +323,8 @@ on_app_state_changed (ShellWindowTracker *tracker,
 
   running = shell_app_get_state (app) == SHELL_APP_STATE_RUNNING;
 
-  usage->last_seen = get_time ();
+  if (running)
+    usage->last_seen = get_time ();
 }
 
 static void
