@@ -2,38 +2,41 @@
 
 #include "config.h"
 
-#include "shell-global-private.h"
+#include <dirent.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <math.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+
+#include <X11/extensions/Xfixes.h>
+#include <canberra.h>
+#include <clutter/glx/clutter-glx.h>
+#include <clutter/x11/clutter-x11.h>
+#include <dbus/dbus-glib.h>
+#include <gdk/gdkx.h>
+#include <gio/gio.h>
+#include <gjs/gjs-module.h>
+#include <girepository.h>
+#include <libxml/parser.h>
+#include <libxml/tree.h>
+#include <libxml/xmlmemory.h>
+#include <meta/display.h>
+#include <meta/util.h>
+
 #include "shell-enum-types.h"
+#include "shell-global-private.h"
+#include "shell-jsapi-compat-private.h"
+#include "shell-marshal.h"
 #include "shell-perf-log.h"
 #include "shell-window-tracker.h"
 #include "shell-wm.h"
 #include "st.h"
 
-#include "display.h"
-#include "util.h"
-#include <clutter/glx/clutter-glx.h>
-#include <clutter/x11/clutter-x11.h>
-#include <gdk/gdkx.h>
-#include <dirent.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <stdlib.h>
-#include <string.h>
-#include <unistd.h>
-#include <dbus/dbus-glib.h>
-#include <gio/gio.h>
-#include <math.h>
-#include <X11/extensions/Xfixes.h>
-#include <gjs/gjs-module.h>
-#include <canberra.h>
-#ifdef HAVE_SYS_RESOURCE_H
-#include <sys/resource.h>
-#endif
-#include "shell-jsapi-compat-private.h"
-
-#define SHELL_DBUS_SERVICE "org.gnome.Shell"
-#define MAGNIFIER_DBUS_SERVICE "org.gnome.Magnifier"
-
 static void grab_notify (GtkWidget *widget, gboolean is_grab, gpointer user_data);
 
 struct _ShellGlobal {
@@ -70,6 +73,8 @@ struct _ShellGlobal {
 
   /* For sound notifications */
   ca_context *sound_context;
+
+  guint32 xdnd_timestamp;
 };
 
 enum {
@@ -92,8 +97,20 @@ enum {
   PROP_FOCUS_MANAGER,
 };
 
+/* Signals */
+enum
+{
+ XDND_POSITION_CHANGED,
+ XDND_LEAVE,
+ XDND_ENTER,
+ NOTIFY_ERROR,
+ LAST_SIGNAL
+};
+
 G_DEFINE_TYPE(ShellGlobal, shell_global, G_TYPE_OBJECT);
 
+static guint shell_global_signals [LAST_SIGNAL] = { 0 };
+
 static void
 shell_global_set_property(GObject         *object,
                           guint            prop_id,
@@ -189,7 +206,8 @@ static void
 shell_global_init (ShellGlobal *global)
 {
   const char *datadir = g_getenv ("GNOME_SHELL_DATADIR");
-  char *imagedir;
+  const char *shell_js = g_getenv("GNOME_SHELL_JS");
+  char *imagedir, **search_path;
 
   if (!datadir)
     datadir = GNOME_SHELL_DATADIR;
@@ -228,6 +246,15 @@ shell_global_init (ShellGlobal *global)
   ca_context_create (&global->sound_context);
   ca_context_change_props (global->sound_context, CA_PROP_APPLICATION_NAME, PACKAGE_NAME, CA_PROP_APPLICATION_ID, "org.gnome.Shell", NULL);
   ca_context_open (global->sound_context);
+
+  if (!shell_js)
+    shell_js = JSDIR;
+  search_path = g_strsplit (shell_js, ":", -1);
+  global->js_context = g_object_new (GJS_TYPE_CONTEXT,
+                                     "search-path", search_path,
+                                     "js-version", "1.8",
+                                     NULL);
+  g_strfreev (search_path);
 }
 
 static void
@@ -238,6 +265,47 @@ shell_global_class_init (ShellGlobalClass *klass)
   gobject_class->get_property = shell_global_get_property;
   gobject_class->set_property = shell_global_set_property;
 
+  /* Emitted from gnome-shell-plugin.c during event handling */
+  shell_global_signals[XDND_POSITION_CHANGED] =
+      g_signal_new ("xdnd-position-changed",
+                    G_TYPE_FROM_CLASS (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0,
+                    NULL, NULL,
+                    _shell_marshal_VOID__INT_INT,
+                    G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
+
+  /* Emitted from gnome-shell-plugin.c during event handling */
+  shell_global_signals[XDND_LEAVE] =
+      g_signal_new ("xdnd-leave",
+                    G_TYPE_FROM_CLASS (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0,
+                    NULL, NULL,
+                    g_cclosure_marshal_VOID__VOID,
+                    G_TYPE_NONE, 0);
+
+  /* Emitted from gnome-shell-plugin.c during event handling */
+  shell_global_signals[XDND_ENTER] =
+      g_signal_new ("xdnd-enter",
+                    G_TYPE_FROM_CLASS (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0,
+                    NULL, NULL,
+                    g_cclosure_marshal_VOID__VOID,
+                    G_TYPE_NONE, 0);
+
+  shell_global_signals[NOTIFY_ERROR] =
+      g_signal_new ("notify-error",
+                    G_TYPE_FROM_CLASS (klass),
+                    G_SIGNAL_RUN_LAST,
+                    0,
+                    NULL, NULL,
+                    gi_cclosure_marshal_generic,
+                    G_TYPE_NONE, 2,
+                    G_TYPE_STRING,
+                    G_TYPE_STRING);
+
   g_object_class_install_property (gobject_class,
                                    PROP_OVERLAY_GROUP,
                                    g_param_spec_object ("overlay-group",
@@ -747,11 +815,10 @@ _shell_global_set_plugin (ShellGlobal *global,
   global->focus_manager = st_focus_manager_get_for_stage (CLUTTER_STAGE (stage));
 }
 
-void
-_shell_global_set_gjs_context (ShellGlobal *global,
-                               GjsContext  *context)
+GjsContext *
+_shell_global_get_gjs_context (ShellGlobal *global)
 {
-  global->js_context = context;
+  return global->js_context;
 }
 
 /**
@@ -817,6 +884,61 @@ shell_global_display_is_grabbed (ShellGlobal *global)
   return meta_display_get_grab_op (display) != META_GRAB_OP_NONE;
 }
 
+/**
+ * shell_global_create_pointer_barrier
+ * @global: a #ShellGlobal
+ * @x1: left X coordinate
+ * @y1: top Y coordinate
+ * @x2: right X coordinate
+ * @y2: bottom Y coordinate
+ * @directions: The directions we're allowed to pass through
+ *
+ * If supported by X creates a pointer barrier.
+ *
+ * Return value: value you can pass to shell_global_destroy_pointer_barrier()
+ */
+guint32
+shell_global_create_pointer_barrier (ShellGlobal *global,
+                                     int x1, int y1, int x2, int y2,
+                                     int directions)
+{
+#if XFIXES_MAJOR >= 5
+  Display *xdpy;
+
+  xdpy = meta_plugin_get_xdisplay (global->plugin);
+
+  return (guint32)
+    XFixesCreatePointerBarrier (xdpy, DefaultRootWindow(xdpy),
+                                x1, y1,
+                                x2, y2,
+                                directions,
+                                0, NULL);
+#else
+  return 0;
+#endif
+}
+
+/**
+ * shell_global_destroy_pointer_barrier
+ * @global: a #ShellGlobal
+ * @barrier: a pointer barrier
+ *
+ * Destroys the @barrier created by shell_global_create_pointer_barrier().
+ */
+void
+shell_global_destroy_pointer_barrier (ShellGlobal *global, guint32 barrier)
+{
+#if XFIXES_MAJOR >= 5
+  Display *xdpy;
+
+  g_return_if_fail (barrier > 0);
+
+  xdpy = meta_plugin_get_xdisplay (global->plugin);
+  XFixesDestroyPointerBarrier (xdpy, (PointerBarrier)barrier);
+#endif
+}
+
+
 /**
  * shell_global_add_extension_importer:
  * @target_object_script: JavaScript code evaluating to a target object
@@ -838,7 +960,6 @@ shell_global_add_extension_importer (ShellGlobal *global,
                                      GError     **error)
 {
   jsval target_object;
-  JSObject *importer;
   JSContext *context = gjs_context_get_native_context (global->js_context);
   char *search_path[2] = { 0, 0 };
 
@@ -873,7 +994,7 @@ shell_global_add_extension_importer (ShellGlobal *global,
     }
 
   search_path[0] = (char*)directory;
-  importer = gjs_define_importer (context, JSVAL_TO_OBJECT (target_object), target_property, (const char **)search_path, FALSE);
+  gjs_define_importer (context, JSVAL_TO_OBJECT (target_object), target_property, (const char **)search_path, FALSE);
   JS_EndRequest (context);
   return TRUE;
  out_error:
@@ -1026,6 +1147,136 @@ shell_global_breakpoint (ShellGlobal *global)
   G_BREAKPOINT ();
 }
 
+/**
+ * shell_global_parse_search_provider:
+ * @global: A #ShellGlobal
+ * @data: description of provider
+ * @name: (out): location to store a display name
+ * @url: (out): location to store template of url
+ * @langs: (out) (transfer full) (element-type utf8): list of supported languages
+ * @icon_data_uri: (out): location to store uri
+ * @error: location to store GError
+ *
+ * Returns: %TRUE on success
+ */
+gboolean
+shell_global_parse_search_provider (ShellGlobal   *global,
+                                    const char    *data,
+                                    char         **name,
+                                    char         **url,
+                                    GList        **langs,
+                                    char         **icon_data_uri,
+                                    GError       **error)
+{
+  xmlDocPtr doc = xmlParseMemory (data, strlen(data));
+  xmlNode *root;
+
+  *name = NULL;
+  *url = NULL;
+  *icon_data_uri = NULL;
+  *langs = NULL;
+
+  if (!doc)
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Malformed xml");
+      return FALSE;
+    }
+
+  root = xmlDocGetRootElement (doc);
+  if (root && root->name && xmlStrcmp (root->name, (const xmlChar *)"OpenSearchDescription") == 0)
+    {
+      xmlNode *child;
+      for (child = root->children; child; child = child->next)
+        {
+            if (!child->name)
+              continue;
+            if (xmlStrcmp (child->name, (const xmlChar *)"Language") == 0)
+              {
+                xmlChar *val = xmlNodeListGetString(doc, child->xmlChildrenNode, 1);
+                if (!val)
+                  continue;
+                *langs = g_list_append (*langs, g_strdup ((char *)val));
+                xmlFree (val);
+              }
+            if (!*name && xmlStrcmp (child->name, (const xmlChar *)"ShortName") == 0)
+              {
+                xmlChar *val = xmlNodeListGetString(doc, child->xmlChildrenNode, 1);
+                *name = g_strdup ((char *)val);
+                xmlFree (val);
+              }
+            if (!*icon_data_uri && xmlStrcmp (child->name, (const xmlChar *)"Image") == 0)
+              {
+                xmlChar *val = xmlNodeListGetString(doc, child->xmlChildrenNode, 1);
+                if (val)
+                  *icon_data_uri = g_strdup ((char *)val);
+                xmlFree (val);
+              }
+            if (!*url && xmlStrcmp (child->name, (const xmlChar *)"Url") == 0)
+              {
+                xmlChar *template;
+                xmlChar *type;
+
+                type = xmlGetProp(child, (const xmlChar *)"type");
+                if (!type)
+                  continue;
+
+                if (xmlStrcmp (type, (const xmlChar *)"text/html") != 0)
+                  {
+                    xmlFree (type);
+                    continue;
+                  }
+                xmlFree (type);
+
+                template = xmlGetProp(child, (const xmlChar *)"template");
+                if (!template)
+                  continue;
+                *url = g_strdup ((char *)template);
+                xmlFree (template);
+              }
+        }
+    }
+  else
+    {
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED, "Invalid OpenSearch document");
+      xmlFreeDoc (doc);
+      return FALSE;
+    }
+  xmlFreeDoc (doc);
+  if (*icon_data_uri && *name && *url)
+    return TRUE;
+
+  if (*icon_data_uri)
+    g_free (*icon_data_uri);
+  else
+    g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                 "search provider doesn't have icon");
+
+  if (*name)
+    g_free (*name);
+  else if (error && !*error)
+    g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                 "search provider doesn't have ShortName");
+
+  if (*url)
+    g_free (*url);
+  else if (error && !*error)
+    g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+                 "search provider doesn't have template for url");
+
+  if (*langs)
+    {
+      g_list_foreach (*langs, (GFunc)g_free, NULL);
+      g_list_free (*langs);
+    }
+
+  *url = NULL;
+  *name = NULL;
+  *icon_data_uri = NULL;
+  *langs = NULL;
+
+  return FALSE;
+}
+
 /**
  * shell_global_gc:
  * @global: A #ShellGlobal
@@ -1052,71 +1303,26 @@ shell_global_gc (ShellGlobal *global)
 void
 shell_global_maybe_gc (ShellGlobal *global)
 {
-  JSContext *context = gjs_context_get_native_context (global->js_context);
-
-  JS_MaybeGC (context);
+  gjs_context_maybe_gc (global->js_context);
 }
 
+/**
+ * shell_global_notify_error:
+ * @global: a #ShellGlobal
+ * @msg: Error message
+ * @details: Error details
+ *
+ * Show a system error notification.  Use this function
+ * when a user-initiated action results in a non-fatal problem
+ * from causes that may not be under system control.  For
+ * example, an application crash.
+ */
 void
-shell_global_grab_dbus_service (ShellGlobal *global)
+shell_global_notify_error (ShellGlobal  *global,
+                           const char   *msg,
+                           const char   *details)
 {
-  GError *error = NULL;
-  DBusGConnection *session;
-  DBusGProxy *bus;
-  guint32 request_name_result;
-  
-  session = dbus_g_bus_get (DBUS_BUS_SESSION, NULL);
-  
-  bus = dbus_g_proxy_new_for_name (session,
-                                   DBUS_SERVICE_DBUS,
-                                   DBUS_PATH_DBUS,
-                                   DBUS_INTERFACE_DBUS);
-  
-  if (!dbus_g_proxy_call (bus, "RequestName", &error,
-                          G_TYPE_STRING, SHELL_DBUS_SERVICE,
-                          G_TYPE_UINT, 0,
-                          G_TYPE_INVALID,
-                          G_TYPE_UINT, &request_name_result,
-                          G_TYPE_INVALID)) 
-    {
-      g_print ("failed to acquire org.gnome.Shell: %s\n", error->message);
-      /* If we somehow got started again, it's not an error to be running
-       * already.  So just exit 0.
-       */
-      exit (0);  
-    }
-
-  /* Also grab org.gnome.Panel to replace any existing panel process,
-   * unless a special environment variable is passed.  The environment
-   * variable is used by the gnome-shell (no --replace) launcher in
-   * Xephyr */
-  if (!g_getenv ("GNOME_SHELL_NO_REPLACE"))
-    {
-      if (!dbus_g_proxy_call (bus, "RequestName", &error, G_TYPE_STRING,
-                              "org.gnome.Panel", G_TYPE_UINT,
-                              DBUS_NAME_FLAG_REPLACE_EXISTING | DBUS_NAME_FLAG_DO_NOT_QUEUE,
-                              G_TYPE_INVALID, G_TYPE_UINT,
-                              &request_name_result, G_TYPE_INVALID))
-        {
-          g_print ("failed to acquire org.gnome.Panel: %s\n", error->message);
-          exit (1);
-        }
-    }
-
-  /* ...and the org.gnome.Magnifier service.
-   */
-  if (!dbus_g_proxy_call (bus, "RequestName", &error,
-                          G_TYPE_STRING, MAGNIFIER_DBUS_SERVICE,
-                          G_TYPE_UINT, 0,
-                          G_TYPE_INVALID,
-                          G_TYPE_UINT, &request_name_result,
-                          G_TYPE_INVALID))
-    {
-      g_print ("failed to acquire %s: %s\n", MAGNIFIER_DBUS_SERVICE, error->message);
-      /* Failing to acquire the magnifer service is not fatal.  Log the error,
-       * but keep going. */
-    }
-  g_object_unref (bus);
+  g_signal_emit_by_name (global, "notify-error", msg, details);
 }
 
 static void
@@ -1130,6 +1336,39 @@ grab_notify (GtkWidget *widget, gboolean was_grabbed, gpointer user_data)
   shell_global_set_stage_input_mode (global, global->input_mode);
 }
 
+/**
+ * shell_global_init_xdnd:
+ * @global: the #ShellGlobal
+ *
+ * Enables tracking of Xdnd events
+ */
+void shell_global_init_xdnd (ShellGlobal *global)
+{
+  long xdnd_version = 5;
+
+  MetaScreen *screen = shell_global_get_screen (global);
+  Window output_window = meta_get_overlay_window (screen);
+
+  MetaDisplay *display = meta_screen_get_display (screen);
+  Display *xdisplay = meta_display_get_xdisplay (display);
+
+  ClutterStage *stage = CLUTTER_STAGE(meta_plugin_get_stage (global->plugin));
+  Window stage_win = clutter_x11_get_stage_window (stage);
+
+  XChangeProperty (xdisplay, stage_win, gdk_x11_get_xatom_by_name ("XdndAware"), XA_ATOM,
+                  32, PropModeReplace, (const unsigned char *)&xdnd_version, 1);
+
+  XChangeProperty (xdisplay, output_window, gdk_x11_get_xatom_by_name ("XdndProxy"), XA_WINDOW,
+                  32, PropModeReplace, (const unsigned char *)&stage_win, 1);
+
+  /*
+   * XdndProxy is additionally set on the proxy window as verification that the
+   * XdndProxy property on the target window isn't a left-over
+   */
+  XChangeProperty (xdisplay, stage_win, gdk_x11_get_xatom_by_name ("XdndProxy"), XA_WINDOW,
+                  32, PropModeReplace, (const unsigned char *)&stage_win, 1);
+}
+
 /**
  * shell_global_format_time_relative_pretty:
  * @global:
@@ -1216,48 +1455,33 @@ shell_global_get_monitors (ShellGlobal *global)
 MetaRectangle *
 shell_global_get_primary_monitor (ShellGlobal  *global)
 {
-  GdkScreen *screen = shell_global_get_gdk_screen (global);
-  GdkRectangle gdk_rect;
+  MetaScreen *screen = shell_global_get_screen (global);
   MetaRectangle rect;
   gint primary = 0;
 
-  /* gdk_screen_get_primary_monitor is only present in gtk-2.20+
-   * and is in a useable state (supports heuristics and fallback modes)
-   * starting with 2.20.1
-   */
-#if !GTK_CHECK_VERSION (2, 20, 1)
-  gint i;
-  gchar *output_name = NULL;
-  gint num_monitors = gdk_screen_get_n_monitors (screen);
-
-  for (i = 0; i < num_monitors; i++)
-    {
-      /* Prefer the laptop's internal screen if present */
-      output_name = gdk_screen_get_monitor_plug_name (screen, i);
-      if (output_name)
-        {
-          gboolean is_lvds = g_ascii_strncasecmp (output_name, "LVDS", 4) == 0;
-          g_free (output_name);
-          if (is_lvds)
-            {
-              primary = i;
-              break;
-            }
-        }
-    }
-#else
-  primary = gdk_screen_get_primary_monitor (screen);
-#endif
-
-  gdk_screen_get_monitor_geometry (screen, primary, &gdk_rect);
-  rect.x = gdk_rect.x;
-  rect.y = gdk_rect.y;
-  rect.width = gdk_rect.width;
-  rect.height = gdk_rect.height;
+  primary = meta_screen_get_primary_monitor (screen);
+  meta_screen_get_monitor_geometry (screen, primary, &rect);
 
   return meta_rectangle_copy (&rect);
 }
 
+/**
+ * shell_global_get_primary_monitor_index:
+ * @global: the #ShellGlobal
+ *
+ * Gets the index of the primary monitor (the one that the
+ * panel is on).
+ *
+ * Return value: the index of the primary monitor
+ */
+int
+shell_global_get_primary_monitor_index (ShellGlobal  *global)
+{
+  MetaScreen *screen = shell_global_get_screen (global);
+
+  return meta_screen_get_primary_monitor (screen);
+}
+
 /**
  * shell_global_get_focus_monitor:
  * @global: the #ShellGlobal
@@ -1321,6 +1545,48 @@ shell_global_get_pointer (ShellGlobal         *global,
   *mods = raw_mods & GDK_MODIFIER_MASK;
 }
 
+/**
+ * shell_global_sync_pointer:
+ * @global: the #ShellGlobal
+ *
+ * Ensures that clutter is aware of the current pointer position,
+ * causing enter and leave events to be emitted if the pointer moved
+ * behind our back (ie, during a pointer grab).
+ */
+void
+shell_global_sync_pointer (ShellGlobal *global)
+{
+  int x, y;
+  GdkModifierType mods;
+  ClutterMotionEvent event;
+
+  gdk_display_get_pointer (gdk_display_get_default (), NULL, &x, &y, &mods);
+
+  event.type = CLUTTER_MOTION;
+  event.time = shell_global_get_current_time (global);
+  event.flags = 0;
+  /* This is wrong: we should be setting event.stage to NULL if the
+   * pointer is not inside the bounds of the stage given the current
+   * stage_input_mode. For our current purposes however, this works.
+   */
+  event.stage = CLUTTER_STAGE (meta_plugin_get_stage (global->plugin));
+  event.x = x;
+  event.y = y;
+  event.modifier_state = mods;
+  event.axes = NULL;
+  event.device = clutter_device_manager_get_core_device (clutter_device_manager_get_default (),
+                                                         CLUTTER_POINTER_DEVICE);
+
+  /* Leaving event.source NULL will force clutter to look it up, which
+   * will generate enter/leave events as a side effect, if they are
+   * needed. We need a better way to do this though... see
+   * http://bugzilla.clutter-project.org/show_bug.cgi?id=2615.
+   */
+  event.source = NULL;
+
+  clutter_event_put ((ClutterEvent *)&event);
+}
+
 /**
  * shell_get_event_state:
  * @event: a #ClutterEvent
@@ -1421,6 +1687,11 @@ shell_global_get_current_time (ShellGlobal *global)
 {
   guint32 time;
   MetaDisplay *display;
+  const ClutterEvent *clutter_event;
+
+  /* In case we have a xdnd timestamp use it */
+  if (global->xdnd_timestamp != 0)
+    return global->xdnd_timestamp;
 
   /* meta_display_get_current_time() will return the correct time
      when handling an X or Gdk event, but will return CurrentTime
@@ -1439,8 +1710,16 @@ shell_global_get_current_time (ShellGlobal *global)
   time = meta_display_get_current_time (display);
   if (time != CLUTTER_CURRENT_TIME)
       return time;
+  /*
+   * We don't use clutter_get_current_event_time as it can give us a
+   * too old timestamp if there is no current event.
+   */
+  clutter_event = clutter_get_current_event ();
 
-  return clutter_get_current_event_time ();
+  if (clutter_event != NULL)
+    return clutter_event_get_time (clutter_event);
+  else
+    return CLUTTER_CURRENT_TIME;
 }
 
 /**
@@ -1678,6 +1957,7 @@ shell_global_run_at_leisure (ShellGlobal         *global,
 /**
  * shell_global_play_theme_sound:
  * @global: the #ShellGlobal
+ * @id: an id, used to cancel later (0 if not needed)
  * @name: the sound name
  *
  * Plays a simple sound picked according to Freedesktop sound theme.
@@ -1685,7 +1965,267 @@ shell_global_run_at_leisure (ShellGlobal         *global,
  */
 void
 shell_global_play_theme_sound (ShellGlobal *global,
+                               guint        id,
                                const char  *name)
 {
-  ca_context_play (global->sound_context, 0, CA_PROP_EVENT_ID, name, NULL);
+  ca_context_play (global->sound_context, id, CA_PROP_EVENT_ID, name, NULL);
+}
+
+/**
+ * shell_global_cancel_theme_sound:
+ * @global: the #ShellGlobal
+ * @id: the id previously passed to shell_global_play_theme_sound()
+ *
+ * Cancels a sound notification.
+ */
+void
+shell_global_cancel_theme_sound (ShellGlobal *global,
+                                 guint id)
+{
+  ca_context_cancel (global->sound_context, id);
+}
+
+/*
+ * Process Xdnd events
+ *
+ * We pass the position and leave events to JS via a signal
+ * where the actual drag & drop handling happens.
+ *
+ * http://www.freedesktop.org/wiki/Specifications/XDND
+ */
+gboolean _shell_global_check_xdnd_event (ShellGlobal  *global,
+                                         XEvent       *xev)
+{
+  MetaScreen *screen = meta_plugin_get_screen (global->plugin);
+  Window output_window = meta_get_overlay_window (screen);
+  MetaDisplay *display = meta_screen_get_display (screen);
+  Display *xdisplay = meta_display_get_xdisplay (display);
+
+  ClutterStage *stage = CLUTTER_STAGE (meta_plugin_get_stage (global->plugin));
+  Window stage_win = clutter_x11_get_stage_window (stage);
+
+  if (xev->xany.window != output_window && xev->xany.window != stage_win)
+    return FALSE;
+
+  if (xev->xany.type == ClientMessage && xev->xclient.message_type == gdk_x11_get_xatom_by_name ("XdndPosition"))
+    {
+      XEvent xevent;
+      Window src = xev->xclient.data.l[0];
+
+      memset (&xevent, 0, sizeof(xevent));
+      xevent.xany.type = ClientMessage;
+      xevent.xany.display = xdisplay;
+      xevent.xclient.window = src;
+      xevent.xclient.message_type = gdk_x11_get_xatom_by_name ("XdndStatus");
+      xevent.xclient.format = 32;
+      xevent.xclient.data.l[0] = output_window;
+      /* flags: bit 0: will we accept the drop? bit 1: do we want more position messages */
+      xevent.xclient.data.l[1] = 2;
+      xevent.xclient.data.l[4] = None;
+
+      XSendEvent (xdisplay, src, False, 0, &xevent);
+
+      /* Store the timestamp of the xdnd position event */
+      global->xdnd_timestamp = xev->xclient.data.l[3];
+      g_signal_emit_by_name (G_OBJECT (global), "xdnd-position-changed",
+                            (int)(xev->xclient.data.l[2] >> 16), (int)(xev->xclient.data.l[2] & 0xFFFF));
+      global->xdnd_timestamp = 0;
+
+      return TRUE;
+    }
+   else if (xev->xany.type == ClientMessage && xev->xclient.message_type == gdk_x11_get_xatom_by_name ("XdndLeave"))
+    {
+      g_signal_emit_by_name (G_OBJECT (global), "xdnd-leave");
+
+      return TRUE;
+    }
+   else if (xev->xany.type == ClientMessage && xev->xclient.message_type == gdk_x11_get_xatom_by_name ("XdndEnter"))
+    {
+      g_signal_emit_by_name (G_OBJECT (global), "xdnd-enter");
+
+      return TRUE;
+    }
+
+    return FALSE;
+}
+
+/**
+ * ShellGetTpContactCb:
+ * @connection: The connection
+ * @contacts: (element-type TelepathyGLib.Contact): List of contacts
+ * @failed: Array of failed contacts
+ */
+
+static void
+shell_global_get_tp_contacts_cb (TpConnection *self,
+                                 guint n_contacts,
+                                 TpContact * const *contacts,
+                                 guint n_failed,
+                                 const TpHandle *failed,
+                                 const GError *error,
+                                 gpointer user_data,
+                                 GObject *weak_object)
+{
+  int i;
+  GList *contact_list = NULL;
+  for (i = 0; i < n_contacts; i++) {
+      contact_list = g_list_append(contact_list, contacts[i]);
+  }
+
+  TpHandle *failed_list = g_new0 (TpHandle, n_failed + 1);
+  memcpy(failed_list, failed, n_failed);
+
+  ((ShellGetTpContactCb)user_data)(self, contact_list, failed_list);
+}
+
+
+/**
+ * shell_get_tp_contacts:
+ * @self: A connection, which must be ready
+ * @n_handles: Number of handles in handles
+ * @handles: (array length=n_handles) (element-type uint): Array of handles
+ * @n_features: Number of features in features
+ * @features: (array length=n_features) (allow-none) (element-type uint):
+ *  Array of features
+ * @callback: (scope async): User callback to run when the contacts are ready
+ *
+ * Wrap tp_connection_get_contacts_by_handle so we can transform the array
+ * into a null-terminated one, which gjs can handle.
+ * We send the original callback to tp_connection_get_contacts_by_handle as
+ * user_data, and we have our own function as callback, which does the
+ * transforming.
+ */
+void
+shell_get_tp_contacts (TpConnection *self,
+                       guint n_handles,
+                       const TpHandle *handles,
+                       guint n_features,
+                       const TpContactFeature *features,
+                       ShellGetTpContactCb callback)
+{
+  tp_connection_get_contacts_by_handle(self, n_handles, handles,
+                                       n_features, features,
+                                       shell_global_get_tp_contacts_cb,
+                                       callback, NULL, NULL);
+}
+
+/**
+ * shell_global_launch_calendar_server:
+ * @global: The #ShellGlobal.
+ *
+ * Launch the gnome-shell-calendar-server helper.
+ */
+void
+shell_global_launch_calendar_server (ShellGlobal *global)
+{
+  const gchar *bin_dir;
+  gchar *calendar_server_exe;
+  GError *error;
+  gchar *argv[2];
+  gint child_standard_input;
+
+  /* launch calendar-server */
+  bin_dir = g_getenv ("GNOME_SHELL_BINDIR");
+  if (bin_dir != NULL)
+    calendar_server_exe = g_strdup_printf ("%s/gnome-shell-calendar-server", bin_dir);
+  else
+    calendar_server_exe = g_strdup_printf (GNOME_SHELL_LIBEXECDIR "/gnome-shell-calendar-server");
+
+  argv[0] = calendar_server_exe;
+  argv[1] = NULL;
+  error = NULL;
+  if (!g_spawn_async_with_pipes (NULL, /* working_directory */
+                                 argv,
+                                 NULL, /* envp */
+                                 0, /* GSpawnFlags */
+                                 NULL, /* child_setup */
+                                 NULL, /* user_data */
+                                 NULL, /* GPid *child_pid */
+                                 &child_standard_input,
+                                 NULL, /* gint *stdout */
+                                 NULL, /* gint *stderr */
+                                 &error))
+    {
+      g_warning ("Error launching `%s': %s (%s %d)",
+                 calendar_server_exe,
+                 error->message,
+                 g_quark_to_string (error->domain),
+                 error->code);
+      g_error_free (error);
+    }
+  /* Note that gnome-shell-calendar-server exits whenever its stdin
+   * file descriptor is HUP'ed. This means that whenever the the shell
+   * process exits or is being replaced, the calendar server is also
+   * exits...and if the shell is being replaced, a new copy of the
+   * calendar server is launched...
+   */
+
+  g_free (calendar_server_exe);
+}
+
+static void
+shell_global_get_self_contact_features_cb (TpConnection *connection,
+                                           guint n_contacts,
+                                           TpContact * const *contacts,
+                                           const GError *error,
+                                           gpointer user_data,
+                                           GObject *weak_object)
+{
+  if (error != NULL) {
+    g_print ("Failed to upgrade self contact: %s", error->message);
+    return;
+  }
+  ((ShellGetSelfContactFeaturesCb)user_data)(connection, *contacts);
+}
+
+/**
+ * shell_get_self_contact_features:
+ * @self: A connection, which must be ready
+ * @n_features: Number of features in features
+ * @features: (array length=n_features) (allow-none) (element-type uint):
+ *  Array of features
+ * @callback: (scope async): User callback to run when the contact is ready
+ * 
+ * Wrap tp_connection_upgrade_contacts due to the lack of support for
+ * proper arrays arguments in GJS.
+ */
+void
+shell_get_self_contact_features (TpConnection *self,
+                                 guint n_features,
+                                 const TpContactFeature *features,
+                                 ShellGetSelfContactFeaturesCb callback)
+{
+  TpContact *self_contact = tp_connection_get_self_contact (self);
+
+  tp_connection_upgrade_contacts (self, 1, &self_contact,
+                                  n_features, features,
+                                  shell_global_get_self_contact_features_cb,
+                                  callback, NULL, NULL);
+}
+
+/**
+ * shell_get_contact_events:
+ * @log_manager: A #TplLogManager
+ * @account: A #TpAccount
+ * @entity: A #TplEntity
+ * @num_events: The number of events to retrieve
+ * @callback: (scope async): User callback to run when the contact is ready
+ * 
+ * Wrap tpl_log_manager_get_filtered_events_async because gjs cannot support
+ * multiple callbacks in the same function call.
+ */
+void
+shell_get_contact_events (TplLogManager *log_manager,
+                          TpAccount *account,
+                          TplEntity *entity,
+                          guint num_events,
+                          GAsyncReadyCallback callback)
+{
+  tpl_log_manager_get_filtered_events_async (log_manager,
+                                             account,
+                                             entity,
+                                             TPL_EVENT_MASK_TEXT,
+                                             num_events,
+                                             NULL, NULL,
+                                             callback, NULL);
 }
