@@ -35,6 +35,7 @@ struct _StContainerPrivate
   GList *children;
   ClutterActor *first_child;
   ClutterActor *last_child;
+  gboolean block_update_pseude_classes;
 };
 
 static void clutter_container_iface_init (ClutterContainerIface *iface);
@@ -50,6 +51,9 @@ st_container_update_pseudo_classes (StContainer *container)
   ClutterActor *first_child, *last_child;
   StContainerPrivate *priv = container->priv;
 
+  if (priv->block_update_pseude_classes)
+    return;
+
   first_item = priv->children;
   first_child = first_item ? first_item->data : NULL;
   if (first_child != priv->first_child)
@@ -91,28 +95,6 @@ st_container_update_pseudo_classes (StContainer *container)
     }
 }
 
-/**
- * st_container_remove_all:
- * @container: An #StContainer
- *
- * Removes all child actors from @container.
- */
-void
-st_container_remove_all (StContainer *container)
-{
-  StContainerPrivate *priv = container->priv;
-
-  /* copied from clutter_group_remove_all() */
-  while (priv->children)
-    {
-      ClutterActor *child = priv->children->data;
-      priv->children = priv->children->next;
-
-      clutter_container_remove_actor (CLUTTER_CONTAINER (container), child);
-    }
-  st_container_update_pseudo_classes (container);
-}
-
 /**
  * st_container_destroy_children:
  * @container: An #StContainer
@@ -124,13 +106,14 @@ st_container_destroy_children (StContainer *container)
 {
   StContainerPrivate *priv = container->priv;
 
-  while (priv->children)
-    {
-      ClutterActor *child = priv->children->data;
-      priv->children = priv->children->next;
+  priv->block_update_pseude_classes = TRUE;
 
-      clutter_actor_destroy (child);
-    }
+  while (priv->children)
+    clutter_actor_destroy (priv->children->data);
+
+  priv->block_update_pseude_classes = FALSE;
+
+  st_container_update_pseudo_classes (container);
 }
 
 void
@@ -159,6 +142,29 @@ st_container_move_child (StContainer  *container,
   clutter_actor_queue_relayout ((ClutterActor*) container);
 }
 
+void
+st_container_move_before (StContainer  *container,
+                          ClutterActor *actor,
+                          ClutterActor *sibling)
+{
+  StContainerPrivate *priv = container->priv;
+  GList *actor_item = NULL;
+  GList *sibling_item = NULL;
+
+  actor_item = g_list_find (priv->children, actor);
+  sibling_item = g_list_find (priv->children, sibling);
+
+  g_return_if_fail (actor_item != NULL);
+  g_return_if_fail (sibling_item != NULL);
+
+  priv->children = g_list_delete_link (priv->children, actor_item);
+  priv->children = g_list_insert_before (priv->children, sibling_item, actor);
+
+  st_container_update_pseudo_classes (container);
+
+  clutter_actor_queue_relayout (CLUTTER_ACTOR (container));
+}
+
 /**
  * st_container_get_children_list:
  * @container: An #StContainer
@@ -439,33 +445,35 @@ filter_by_position (GList            *children,
       clutter_actor_get_allocation_box (child, &cbox);
 
       /* Filter out children if they are in the wrong direction from
-       * @rbox, or if they don't overlap it.
+       * @rbox, or if they don't overlap it. To account for floating-
+       * point imprecision, an actor is "down" (etc.) from an another
+       * actor even if it overlaps it by up to 0.1 pixels.
        */
       switch (direction)
         {
         case GTK_DIR_UP:
-          if (cbox.y2 > rbox->y1)
+          if (cbox.y2 > rbox->y1 + 0.1)
             continue;
           if (cbox.x1 >= rbox->x2 || cbox.x2 <= rbox->x1)
             continue;
           break;
 
         case GTK_DIR_DOWN:
-          if (cbox.y1 < rbox->y2)
+          if (cbox.y1 < rbox->y2 - 0.1)
             continue;
           if (cbox.x1 >= rbox->x2 || cbox.x2 <= rbox->x1)
             continue;
           break;
 
         case GTK_DIR_LEFT:
-          if (cbox.x2 > rbox->x1)
+          if (cbox.x2 > rbox->x1 + 0.1)
             continue;
           if (cbox.y1 >= rbox->y2 || cbox.y2 <= rbox->y1)
             continue;
           break;
 
         case GTK_DIR_RIGHT:
-          if (cbox.x1 < rbox->x2)
+          if (cbox.x1 < rbox->x2 - 0.1)
             continue;
           if (cbox.y1 >= rbox->y2 || cbox.y2 <= rbox->y1)
             continue;
@@ -640,23 +648,9 @@ st_container_navigate_focus (StWidget         *widget,
        * any child is inconsistently scaled, then the focus chain will
        * probably be unpredictable.
        */
-      if (from)
+      if (focus_child)
         {
-          if (from == focus_child)
-            clutter_actor_get_allocation_box (focus_child, &sort_data.box);
-          else
-            {
-              float cx, cy, fx, fy, fw, fh;
-
-              clutter_actor_get_transformed_position (CLUTTER_ACTOR (container), &cx, &cy);
-              clutter_actor_get_transformed_position (from, &fx, &fy);
-              clutter_actor_get_transformed_size (from, &fw, &fh);
-
-              sort_data.box.x1 = fx - cx;
-              sort_data.box.x2 = fx - cx + fw;
-              sort_data.box.y1 = fy - cy;
-              sort_data.box.y2 = fy - cy + fh;
-            }
+          clutter_actor_get_allocation_box (focus_child, &sort_data.box);
         }
       else
         {
