@@ -3,6 +3,7 @@
  * st-tooltip.c: Plain tooltip actor
  *
  * Copyright 2008, 2009 Intel Corporation
+ * Copyright 2009 Red Hat, Inc.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU Lesser General Public License,
@@ -14,12 +15,7 @@
  * more details.
  *
  * You should have received a copy of the GNU Lesser General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
- * Boston, MA 02111-1307, USA.
- *
- * Written by: Thomas Wood <thomas@linux.intel.com>
- *
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
  */
 
 /**
@@ -35,6 +31,7 @@
 #include "config.h"
 #endif
 
+#include <math.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -45,6 +42,7 @@
 #include "st-tooltip.h"
 
 #include "st-widget.h"
+#include "st-label.h"
 #include "st-private.h"
 
 enum
@@ -60,16 +58,23 @@ enum
 
 struct _StTooltipPrivate
 {
-  ClutterActor    *label;
-
-  gfloat           arrow_offset;
-  gboolean         actor_below;
+  StLabel         *label;
 
   ClutterGeometry *tip_area;
 };
 
+extern gfloat st_slow_down_factor;
+
 G_DEFINE_TYPE (StTooltip, st_tooltip, ST_TYPE_WIDGET);
 
+static void st_tooltip_show (ClutterActor *self);
+static void st_tooltip_show_all (ClutterActor *self);
+static void st_tooltip_hide_all (ClutterActor *self);
+
+static void st_tooltip_constrain (StTooltip             *tooltip,
+                                  const ClutterGeometry *geometry,
+                                  ClutterGeometry       *adjusted_geometry);
+
 static void
 st_tooltip_set_property (GObject      *gobject,
                          guint         prop_id,
@@ -104,7 +109,7 @@ st_tooltip_get_property (GObject    *gobject,
   switch (prop_id)
     {
     case PROP_LABEL:
-      g_value_set_string (value, clutter_text_get_text (CLUTTER_TEXT (priv->label)));
+      g_value_set_string (value, st_label_get_text (priv->label));
       break;
 
     case PROP_TIP_AREA:
@@ -117,29 +122,6 @@ st_tooltip_get_property (GObject    *gobject,
     }
 }
 
-static void
-st_tooltip_style_changed (StWidget *self)
-{
-  StTooltipPrivate *priv;
-  StThemeNode *theme_node;
-  ClutterColor color;
-  const PangoFontDescription *font;
-  gchar *font_string;
-
-  priv = ST_TOOLTIP (self)->priv;
-  theme_node = st_widget_get_theme_node (self);
-
-  st_theme_node_get_foreground_color (theme_node, &color);
-  clutter_text_set_color (CLUTTER_TEXT (priv->label), &color);
-
-  font = st_theme_node_get_font (theme_node);
-  font_string = pango_font_description_to_string (font);
-  clutter_text_set_font_name (CLUTTER_TEXT (priv->label), font_string);
-  g_free (font_string);
-
-  ST_WIDGET_CLASS (st_tooltip_parent_class)->style_changed (self);
-}
-
 static void
 st_tooltip_get_preferred_width (ClutterActor *self,
                                 gfloat        for_height,
@@ -148,40 +130,23 @@ st_tooltip_get_preferred_width (ClutterActor *self,
 {
   StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
-  gfloat label_height, arrow_height;
-  ClutterActor *arrow_image;
+  gfloat label_height;
 
   st_theme_node_adjust_for_height (theme_node, &for_height);
 
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-  if (arrow_image)
-    {
-      clutter_actor_get_preferred_height (arrow_image,
-                                          -1,
-                                          NULL,
-                                          &arrow_height);
-    }
-  else
-    {
-      arrow_height = 0;
-    }
-
   if (for_height > -1)
     {
-      label_height = for_height - arrow_height;
+      label_height = for_height;
     }
   else
     {
       label_height = -1;
     }
 
-  if (priv->label)
-    {
-      clutter_actor_get_preferred_width (priv->label,
-                                         label_height,
-                                         min_width_p,
-                                         natural_width_p);
-    }
+  clutter_actor_get_preferred_width (CLUTTER_ACTOR (priv->label),
+                                     label_height,
+                                     min_width_p,
+                                     natural_width_p);
 
   st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
 }
@@ -194,44 +159,19 @@ st_tooltip_get_preferred_height (ClutterActor *self,
 {
   StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
-  gfloat arrow_height;
   gfloat min_label_h, natural_label_h;
-  ClutterActor *arrow_image;
 
   st_theme_node_adjust_for_width (theme_node, &for_width);
 
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-
-  if (arrow_image && !priv->actor_below)
-    {
-      clutter_actor_get_preferred_height (arrow_image,
-                                          -1,
-                                          NULL,
-                                          &arrow_height);
-    }
-  else
-    {
-      arrow_height = 0;
-    }
-
-  if (priv->label)
-    {
-      clutter_actor_get_preferred_height (priv->label,
-                                          for_width,
-                                          &min_label_h,
-                                          &natural_label_h);
-    }
-  else
-    {
-      min_label_h = 0;
-      natural_label_h = 0;
-    }
-
+  clutter_actor_get_preferred_height (CLUTTER_ACTOR (priv->label),
+                                      for_width,
+                                      &min_label_h,
+                                      &natural_label_h);
   if (min_height_p)
-    *min_height_p = arrow_height + min_label_h;
+    *min_height_p = min_label_h;
 
   if (natural_height_p)
-    *natural_height_p = arrow_height + natural_label_h;
+    *natural_height_p = natural_label_h;
 
   st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
 }
@@ -243,9 +183,7 @@ st_tooltip_allocate (ClutterActor          *self,
 {
   StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
   StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (self));
-  ClutterActorBox content_box, child_box, arrow_box;
-  gfloat arrow_height, arrow_width;
-  ClutterActor *border_image, *arrow_image;
+  ClutterActorBox content_box, child_box;
 
   CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->allocate (self,
                                                            box,
@@ -253,100 +191,23 @@ st_tooltip_allocate (ClutterActor          *self,
 
   st_theme_node_get_content_box (theme_node, box, &content_box);
 
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-
-  if (arrow_image && !priv->actor_below)
-    {
-      clutter_actor_get_preferred_height (arrow_image, -1, NULL, &arrow_height);
-      clutter_actor_get_preferred_width (arrow_image, -1, NULL, &arrow_width);
-
-      arrow_box.x1 = (float)(priv->arrow_offset) - (int)(arrow_width / 2);
-      arrow_box.y1 = 0;
-      arrow_box.x2 = arrow_box.x1 + arrow_width;
-      arrow_box.y2 = arrow_box.y1 + arrow_height;
-
-      clutter_actor_allocate (arrow_image, &arrow_box, flags);
-    }
-  else
-    {
-      arrow_height = 0;
-      arrow_width = 0;
-    }
-
   child_box.x1 = child_box.y1 = 0;
   child_box.x2 = (box->x2 - box->x1);
   child_box.y2 = (box->y2 - box->y1);
 
-  /* remove the space that is used by the arrow */
-  child_box.y1 += arrow_height;
+  child_box = content_box;
 
-  border_image = st_widget_get_border_image (ST_WIDGET (self));
-  if (border_image)
-    clutter_actor_allocate (border_image, &child_box, flags);
-
-  if (priv->label)
-    {
-      child_box = content_box;
-      child_box.y1 += arrow_height;
-
-      clutter_actor_allocate (priv->label, &child_box, flags);
-    }
+  clutter_actor_allocate (CLUTTER_ACTOR (priv->label), &child_box, flags);
 }
 
-
 static void
 st_tooltip_paint (ClutterActor *self)
 {
-  ClutterActor *border_image, *arrow_image;
   StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
 
-  border_image = st_widget_get_border_image (ST_WIDGET (self));
-  if (border_image)
-    clutter_actor_paint (border_image);
+  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->paint (self);
 
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-  if (arrow_image && !priv->actor_below)
-    clutter_actor_paint (arrow_image);
-
-  clutter_actor_paint (priv->label);
-}
-
-static void
-st_tooltip_map (ClutterActor *self)
-{
-  StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
-  ClutterActor *border_image, *arrow_image;
-
-  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->map (self);
-
-  border_image = st_widget_get_border_image (ST_WIDGET (self));
-  if (border_image)
-    clutter_actor_map (border_image);
-
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-  if (arrow_image)
-    clutter_actor_map (arrow_image);
-
-  clutter_actor_map (priv->label);
-}
-
-static void
-st_tooltip_unmap (ClutterActor *self)
-{
-  StTooltipPrivate *priv = ST_TOOLTIP (self)->priv;
-  ClutterActor *border_image, *arrow_image;
-
-  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->unmap (self);
-
-  border_image = st_widget_get_border_image (ST_WIDGET (self));
-  if (border_image)
-    clutter_actor_unmap (border_image);
-
-  arrow_image = st_widget_get_background_image (ST_WIDGET (self));
-  if (arrow_image)
-    clutter_actor_unmap (arrow_image);
-
-  clutter_actor_unmap (priv->label);
+  clutter_actor_paint (CLUTTER_ACTOR (priv->label));
 }
 
 static void
@@ -356,7 +217,7 @@ st_tooltip_dispose (GObject *self)
 
   if (priv->label)
     {
-      clutter_actor_destroy (priv->label);
+      clutter_actor_destroy (CLUTTER_ACTOR (priv->label));
       priv->label = NULL;
     }
 
@@ -368,7 +229,6 @@ st_tooltip_class_init (StTooltipClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
-  StWidgetClass *widget_class = ST_WIDGET_CLASS (klass);
   GParamSpec *pspec;
 
   g_type_class_add_private (klass, sizeof (StTooltipPrivate));
@@ -381,10 +241,9 @@ st_tooltip_class_init (StTooltipClass *klass)
   actor_class->get_preferred_height = st_tooltip_get_preferred_height;
   actor_class->allocate = st_tooltip_allocate;
   actor_class->paint = st_tooltip_paint;
-  actor_class->map = st_tooltip_map;
-  actor_class->unmap = st_tooltip_unmap;
-
-  widget_class->style_changed = st_tooltip_style_changed;
+  actor_class->show = st_tooltip_show;
+  actor_class->show_all = st_tooltip_show_all;
+  actor_class->hide_all = st_tooltip_hide_all;
 
   pspec = g_param_spec_string ("label",
                                "Label",
@@ -405,11 +264,7 @@ st_tooltip_init (StTooltip *tooltip)
 {
   tooltip->priv = ST_TOOLTIP_GET_PRIVATE (tooltip);
 
-  tooltip->priv->label = g_object_new (CLUTTER_TYPE_TEXT,
-                                       "line-alignment", PANGO_ALIGN_CENTER,
-                                       "ellipsize", PANGO_ELLIPSIZE_END,
-                                       "use-markup", TRUE,
-                                       NULL);
+  tooltip->priv->label = ST_LABEL (st_label_new (NULL));
 
   tooltip->priv->tip_area = NULL;
 
@@ -426,17 +281,14 @@ st_tooltip_update_position (StTooltip *tooltip)
 {
   StTooltipPrivate *priv = tooltip->priv;
   ClutterGeometry *tip_area = tooltip->priv->tip_area;
+  ClutterGeometry geometry;
+  ClutterGeometry adjusted_geometry;
   gfloat tooltip_w, tooltip_h, tooltip_x, tooltip_y;
-  gfloat stage_w, stage_h;
-  ClutterActor *stage;
-
-  /* ensure the tooltip with is not fixed size */
-  clutter_actor_set_size ((ClutterActor*) tooltip, -1, -1);
 
   /* if no area set, just position ourselves top left */
   if (!priv->tip_area)
     {
-      clutter_actor_set_position ((ClutterActor*) tooltip, 0, 0);
+      clutter_actor_set_anchor_point ((ClutterActor*) tooltip, 0, 0);
       return;
     }
 
@@ -451,47 +303,22 @@ st_tooltip_update_position (StTooltip *tooltip)
   tooltip_x = (int)(tip_area->x + (tip_area->width / 2) - (tooltip_w / 2));
   tooltip_y = (int)(tip_area->y + tip_area->height);
 
-  stage = clutter_actor_get_stage ((ClutterActor *) tooltip);
-  if (!stage)
-    {
-      return;
-    }
-  clutter_actor_get_size (stage, &stage_w, &stage_h);
+  geometry.x = tooltip_x;
+  geometry.y = tooltip_y;
+  geometry.width = ceil (tooltip_w);
+  geometry.height = ceil (tooltip_h);
 
-  /* make sure the tooltip is not off screen vertically */
-  if (tooltip_w > stage_w)
-    {
-      tooltip_x = 0;
-      clutter_actor_set_width ((ClutterActor*) tooltip, stage_w);
-    }
-  else if (tooltip_x < 0)
-    {
-      tooltip_x = 0;
-    }
-  else if (tooltip_x + tooltip_w > stage_w)
-    {
-      tooltip_x = (int)(stage_w) - tooltip_w;
-    }
+  st_tooltip_constrain (tooltip, &geometry, &adjusted_geometry);
 
-  /* make sure the tooltip is not off screen horizontally */
-  if (tooltip_y + tooltip_h > stage_h)
-    {
-      priv->actor_below = TRUE;
+  tooltip_x = adjusted_geometry.x;
+  tooltip_y = adjusted_geometry.y;
 
-      /* re-query size as may have changed */
-      clutter_actor_get_preferred_height ((ClutterActor*) tooltip,
-                                          -1, NULL, &tooltip_h);
-      tooltip_y = tip_area->y - tooltip_h;
-    }
-  else
-    {
-      priv->actor_below = FALSE;
-    }
-
-  /* calculate the arrow offset */
-  priv->arrow_offset = tip_area->x + tip_area->width / 2 - tooltip_x;
-
-  clutter_actor_set_position ((ClutterActor*) tooltip, tooltip_x, tooltip_y);
+  /* Since we are updating the position out of st_widget_allocate(), we can't
+   * call clutter_actor_set_position(), since that would trigger another
+   * allocation cycle. Instead, we adjust the anchor position which moves
+   * the tooltip actor on the screen without changing its allocation
+   */
+  clutter_actor_set_anchor_point ((ClutterActor*) tooltip, -tooltip_x, -tooltip_y);
 }
 
 /**
@@ -502,12 +329,12 @@ st_tooltip_update_position (StTooltip *tooltip)
  *
  * Returns: the text for the tooltip. This must not be freed by the application
  */
-G_CONST_RETURN gchar *
+const gchar *
 st_tooltip_get_label (StTooltip *tooltip)
 {
   g_return_val_if_fail (ST_IS_TOOLTIP (tooltip), NULL);
 
-  return clutter_text_get_text (CLUTTER_TEXT (tooltip->priv->label));
+  return st_label_get_text (tooltip->priv->label);
 }
 
 /**
@@ -527,113 +354,36 @@ st_tooltip_set_label (StTooltip   *tooltip,
 
   priv = tooltip->priv;
 
-  clutter_text_set_text (CLUTTER_TEXT (priv->label), text);
+  st_label_set_text (priv->label, text);
 
   g_object_notify (G_OBJECT (tooltip), "label");
 }
 
-/**
- * st_tooltip_show:
- * @tooltip: a #StTooltip
- *
- * Show the tooltip relative to the associated widget.
- */
-void
-st_tooltip_show (StTooltip *tooltip)
+static void
+st_tooltip_show (ClutterActor *self)
 {
-  StTooltipPrivate *priv;
-  ClutterActor *parent;
-  ClutterActor *stage;
-  ClutterActor *self = CLUTTER_ACTOR (tooltip);
-  ClutterAnimation *animation;
-
-  /* make sure we're not currently already animating (e.g. hiding) */
-  animation = clutter_actor_get_animation (CLUTTER_ACTOR (tooltip));
-  if (animation)
-    clutter_animation_completed (animation);
-
-  priv = tooltip->priv;
-  parent = clutter_actor_get_parent (self);
-  stage = clutter_actor_get_stage (self);
-
-  if (!stage)
-    {
-      g_warning ("StTooltip is not on any stage.");
-      return;
-    }
-
-  /* make sure we're parented on the stage */
-  if (G_UNLIKELY (parent != stage))
-    {
-      g_object_ref (self);
-      clutter_actor_unparent (self);
-      clutter_actor_set_parent (self, stage);
-      g_object_unref (self);
-      parent = stage;
-    }
-
-  /* raise the tooltip to the top */
-  clutter_container_raise_child (CLUTTER_CONTAINER (stage),
-                                 CLUTTER_ACTOR (tooltip),
-                                 NULL);
+  StTooltip *tooltip = ST_TOOLTIP (self);
 
   st_tooltip_update_position (tooltip);
 
   /* finally show the tooltip... */
   CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->show (self);
-
-  /* and give it some bounce! */
-  g_object_set (G_OBJECT (self),
-                "scale-center-x", priv->arrow_offset,
-                "scale-center-y", (priv->actor_below) ? clutter_actor_get_height (self) : 0,
-                NULL);
-  clutter_actor_set_scale (self, 0.0, 0.0);
-  clutter_actor_animate (self, CLUTTER_EASE_OUT_ELASTIC,
-                         500,
-                         "scale-x", 1.0,
-                         "scale-y", 1.0,
-                         NULL);
 }
 
 static void
-st_tooltip_hide_complete (ClutterAnimation *animation,
-                          ClutterActor     *actor)
+st_tooltip_show_all (ClutterActor *self)
 {
-  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->hide (actor);
-  g_signal_handlers_disconnect_by_func (actor,
-                                        st_tooltip_hide_complete,
-                                        actor);
+  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->show_all (self);
+
+  clutter_actor_show_all (CLUTTER_ACTOR (ST_TOOLTIP (self)->priv->label));
 }
 
-/**
- * st_tooltip_hide:
- * @tooltip: a #StTooltip
- *
- * Hide the tooltip
- */
-void
-st_tooltip_hide (StTooltip *tooltip)
+static void
+st_tooltip_hide_all (ClutterActor *self)
 {
-  ClutterAnimation *animation;
+  CLUTTER_ACTOR_CLASS (st_tooltip_parent_class)->hide_all (self);
 
-  g_return_if_fail (ST_TOOLTIP (tooltip));
-
-  /* make sure we're not currently already animating (e.g. hiding) */
-  animation = clutter_actor_get_animation (CLUTTER_ACTOR (tooltip));
-  if (animation)
-    clutter_animation_completed (animation);
-
-  g_object_set (G_OBJECT (tooltip),
-                "scale-center-x", tooltip->priv->arrow_offset,
-                NULL);
-  animation =
-    clutter_actor_animate (CLUTTER_ACTOR (tooltip), CLUTTER_EASE_IN_SINE,
-                           150,
-                           "scale-x", 0.0,
-                           "scale-y", 0.0,
-                           NULL);
-  g_signal_connect (animation, "completed",
-                    G_CALLBACK (st_tooltip_hide_complete), tooltip);
+  clutter_actor_hide_all (CLUTTER_ACTOR (ST_TOOLTIP (self)->priv->label));
 }
 
 /**
@@ -653,7 +403,8 @@ st_tooltip_set_tip_area (StTooltip             *tooltip,
     g_boxed_free (CLUTTER_TYPE_GEOMETRY, tooltip->priv->tip_area);
   tooltip->priv->tip_area = g_boxed_copy (CLUTTER_TYPE_GEOMETRY, area);
 
-  st_tooltip_update_position (tooltip);
+  if (clutter_actor_get_stage (CLUTTER_ACTOR (tooltip)))
+    st_tooltip_update_position (tooltip);
 }
 
 /**
@@ -665,10 +416,109 @@ st_tooltip_set_tip_area (StTooltip             *tooltip,
  * Returns: the #ClutterGeometry, owned by the tooltip which must not be freed
  * by the application.
  */
-G_CONST_RETURN ClutterGeometry*
+const ClutterGeometry*
 st_tooltip_get_tip_area (StTooltip *tooltip)
 {
   g_return_val_if_fail (ST_IS_TOOLTIP (tooltip), NULL);
 
   return tooltip->priv->tip_area;
 }
+
+typedef struct {
+  StTooltipConstrainFunc func;
+  gpointer data;
+  GDestroyNotify notify;
+} ConstrainFuncClosure;
+
+static void
+constrain_func_closure_free (gpointer data)
+{
+  ConstrainFuncClosure *closure = data;
+  if (closure->notify)
+    closure->notify (closure->data);
+  g_slice_free (ConstrainFuncClosure, data);
+}
+
+static GQuark
+st_tooltip_constrain_func_quark (void)
+{
+  static GQuark value = 0;
+  if (G_UNLIKELY (value == 0))
+    value = g_quark_from_static_string ("st-tooltip-constrain-func");
+  return value;
+}
+
+/**
+ * st_tooltip_set_constrain_func:
+ * @stage: a #ClutterStage
+ * @func: (allow-none): function to be called to constrain tooltip position
+ * @data: (allow-none): user data to pass to @func
+ * @notify: (allow-none): function to be called when @data is no longer needed
+ *
+ * Sets a callback function that will be used to constrain the position
+ * of tooltips within @stage. This can be used, for example, if the stage
+ * spans multiple monitors and tooltips should be positioned not to cross
+ * monitors.
+ */
+void
+st_tooltip_set_constrain_func (ClutterStage           *stage,
+                               StTooltipConstrainFunc  func,
+                               gpointer                data,
+                               GDestroyNotify          notify)
+{
+  ConstrainFuncClosure *closure;
+
+  g_return_if_fail (CLUTTER_IS_STAGE (stage));
+
+  if (func)
+    {
+      closure = g_slice_new (ConstrainFuncClosure);
+      closure->func = func;
+      closure->data = data;
+      closure->notify = notify;
+    }
+  else
+    closure = NULL;
+
+  g_object_set_qdata_full (G_OBJECT (stage), st_tooltip_constrain_func_quark (),
+                           closure, constrain_func_closure_free);
+}
+
+static void
+st_tooltip_constrain (StTooltip             *tooltip,
+                      const ClutterGeometry *geometry,
+                      ClutterGeometry       *adjusted_geometry)
+{
+  ConstrainFuncClosure *closure;
+
+  ClutterActor *stage = clutter_actor_get_stage (CLUTTER_ACTOR (tooltip));
+
+  *adjusted_geometry = *geometry;
+
+  if (stage == NULL)
+    return;
+
+  closure = g_object_get_qdata (G_OBJECT (stage), st_tooltip_constrain_func_quark ());
+  if (closure)
+    {
+      closure->func (tooltip, geometry, adjusted_geometry, closure->data);
+    }
+  else
+    {
+      ClutterActor *parent;
+      gfloat parent_w, parent_h;
+
+      parent = clutter_actor_get_parent ((ClutterActor *) tooltip);
+      clutter_actor_get_size (parent, &parent_w, &parent_h);
+
+      /* make sure the tooltip is not off parent horizontally */
+      if (adjusted_geometry->x < 0)
+        adjusted_geometry->x = 0;
+      else if (adjusted_geometry->x + adjusted_geometry->width > parent_w)
+        adjusted_geometry->x = (int)(parent_w) - adjusted_geometry->width;
+
+      /* make sure the tooltip is not off parent vertically */
+      if (adjusted_geometry->y + adjusted_geometry->height > parent_h)
+        adjusted_geometry->y = parent_h - adjusted_geometry->height;
+    }
+}
